
;; routines for loading pgplot libraries for cmucl - would like
;; to put this into into pgplot-cmucl, but cmucl barfs about
;; undefined foreign symbols unless libraries get loaded before
;; they are referenced -- (eval-when (load eval compile)...) does
;; not do the trick, alas

(in-package pgplot)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; to make cpgplot.o
;;  break up libcpgplot.a into
;;   .o files and then recombine them into a single object file cpgplot.o
;;  using the following commands:
;;   # break up libcpgplot.a
;;     ar -x libcpgplot.a
;;   # produce single object file
;;     ld -r *.o -o cpgplot.o
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(if (not (boundp 'cl-user::*dynamic-obj-directory* ))
    (error "The variable cl-user::*dynamic-obj-directory* is unbound
pgplot-cmucl.lisp needs this bound to the directory where it can find
the pgplot object file cpgplot.o"))

(eval-when (load eval compile)
  (if (not (boundp 'cl-user::*pgplot-libraries*))
      (format
       nil
 "NOTICE: you have not bound cl-user::*pgplot-libraries*, so pgplot will
try to guess the appropriate libraries to load along with pgplot.~%~%")))


;; the libraries that need to be loaded to get pgplot to work - note
;; that these worked one one setup, and may not work on other setups
;;
;; if user has specified cl-user::*pgplot-libraries*, use that,
;; else use the following built-in guesses
;;
;; note that pgplot now requires libpng and libz 
;;
(defvar cl-user::*pgplot-libraries* nil) ;; make it NIL if undefined
(if (not cl-user::*pgplot-libraries*)
    (setf cl-user::*pgplot-libraries*
      #+linux '("-lpgplot" "-lc" "-lz"
		"-lpng" "-lcpgplot" "/usr/X11R6/lib/libX11.so.6")
      #+solaris '("-lpng" "-lz" "-lpgplot" "-lc" "-lcpgplot" "-lF77" "-lM77"
		  "-lsunmath")
      #+(not (or solaris linux))
      (error "You must set cl-user::*pgplot-libraries* to the proper value
 for your operating system")         
      ))

 
;; now define *pgplot-libraries* in pgplot package
(defparameter *pgplot-libraries* cl-user::*pgplot-libraries* )


(defvar *pgplot-libraries-loaded* nil)
;; load the library if not loaded
(eval-when (load eval compile)
  (unless *pgplot-libraries-loaded*
    (format t "Loading pgplot libraries...~%")
    (load-foreign (concatenate 'string cl-user::*dynamic-obj-directory*
			       "/"
			       "cpgplot.o")
		  :libraries *pgplot-libraries*)
    (setf *pgplot-libraries-loaded* t)))







