
(in-package pgplot)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CMUCL specific FFI stuff
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; first, bind the functions of the pgplot library to make
;; a set of 'raw' functions, then make a higher level
;; set of functions that use the low level functions


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; define alien routines in library
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def-alien-routine ("cpgarro" pgarro-raw) void
  (x1 float) (y1 float) (x2 float) (y2 float))

(def-alien-routine ("cpgask" pgask-raw) void
  (flag int))

(def-alien-routine ("cpgaxis" pgaxis-raw) void
  (opt char)
  (x1 float)(y1 float)(x2 float)(y2 float)(v1 float)(v2 float)
  (step float)(dmajl float)(dmajr float)(fmin float)(disp float)
  (orient float) (nsub int))

(def-alien-routine ("cpgband" pgband-raw) int
  (mode int) (posn int) (xref float) (yref float)
  (x single-float :out)
  (y single-float :out)
  (ch char :out))

(def-alien-routine ("cpgbbuf" pgbbuf-raw) void)

;; PGBEG absent (use PGCLOS instead)

(def-alien-routine ("cpgbin" pgbin-raw) void
  (nbin int)
  (x    (* single-float) :in)
  (data (* single-float) :in)
  (center int))

(def-alien-routine ("cpgbox" pgbox-raw) void
  (xopt c-string) (xtick float) (nxsub int)
  (yopt c-string) (ytick float) (nysub int))

(def-alien-routine ("cpgcirc" pgcirc-raw) void
  (xcent float) (ycent float) (radius float))

(def-alien-routine ("cpgclos" pgclos-raw) void)

(def-alien-routine ("cpgconb" pgconb-raw) void
  (idim int) (jdim int) (i1 int) (i2 int) (j1 int) (j2 int) (nc int)
  (a     (* single-float) :in) ;; 2d array
  (c     (* single-float) :in) 
  (blank (* single-float) :in) )

(def-alien-routine ("cpgconf" pgconf-raw) void
  (idim int) (jdim int) (i1 int) (i2 int) (j1 int) (j2 int) 
  (a     (* single-float) :in) ;; 2d array
  (c1 float) (c2 float)
  (tr (* single-float) :in) )


(def-alien-routine ("cpgconl" pgconl-raw) void
  (a     (* single-float) :in) ;; 2d array
  (idim int) (jdim int)
  (i1 int) (i2 int) (j1 int) (j2 int)
  (c   single-float :in) ;; just a single number
  (tr (* single-float) :in)
  (label c-string)
  (intval int)
  (minint int))


(def-alien-routine ("cpgcons" pgcons-raw) void
  (a     (* single-float) :in) ;; 2d array
  (idim int) (jdim int)
  (i1 int) (i2 int) (j1 int) (j2 int)
  (c     (* single-float) :in) ;; 1d array
  (nc int)
  (tr (* single-float) :in) )

(def-alien-routine ("cpgcont" pgcont-raw) void
  (a     (* single-float) :in) ;; 2d array
  (idim int) (jdim int)
  (i1 int) (i2 int) (j1 int) (j2 int)
  (c     (* single-float) :in) ;; 1d array
  (nc int)
  (tr (* single-float) :in) )


;; PGCONX absent

(def-alien-routine ("cpgctab" pgctab-raw) void
  (l (* single-float) :in)
  (r (* single-float) :in)
  (g (* single-float) :in)
  (b (* single-float) :in)
  (nc int)
  (contra float)
  (bright float))




(def-alien-routine ("cpgcurs" pgcurs-raw) void
  (x float :in-out)   (y float :in-out)   (c char :out))



(def-alien-routine ("cpgdraw" pgdraw-raw) void
  (x float) (y float))

(def-alien-routine ("cpgebuf" pgebuf-raw) void)

(def-alien-routine ("cpgend" pgeend-raw) void)

(def-alien-routine ("cpgenv" pgenv-raw) int
  (xmin float)
  (xmax float)
  (ymin float)
  (ymax float)
  (just int)
  (axis int))


(def-alien-routine ("cpgeras" pgeras-raw) void)

;; PGERR1 absent


(def-alien-routine ("cpgerrb" pgerrb-raw) void
  (dir int)
  (n int)
  (x (* single-float) :in)
  (y (* single-float) :in)
  (e (* single-float) :in)
  (tt float))
  

;; PGERRX absent
;; PGERRY absent
;; PGETXT absent
;; PGFUNT absent (impossible)
;; PGFUNX absent (impossible)
;; PGFUNY absent (impossible)

(def-alien-routine ("cpggray" pggray-raw) void
  (a     (* single-float) :in) ;; 2d array
  (idim int) (jdim int) (i1 int) (i2 int) (j1 int) (j2 int) 
  (fg float) (bg float)
  (tr (* single-float) :in) )

;; PGHI2D absent

(def-alien-routine ("cpghist" pghist-raw) void
  (n int)
  (data     (* single-float) :in)
  (datmin float) (datmax float)
  (nbin int) (pgflag int))

(def-alien-routine ("cpgiden" pgiden-raw) void)

(def-alien-routine ("cpgimag" pgimag-raw) void
  (a     (* single-float) :in) ;; 2d array
  (idim int) (jdim int) (i1 int) (i2 int) (j1 int) (j2 int) 
  (a1 float) (a2 float)
  (tr (* single-float) :in) )

(def-alien-routine ("cpglab" pglab-raw) int
  (xlabel c-string)
  (ylabel c-string)
  (toplabel c-string))


;; PGLCUR absent

(def-alien-routine ("cpgldev" pgeldev-raw) void)


(def-alien-routine ("cpglen" pglen-raw) void
  (units int) (string c-string)
  (xl float :out) (yl float :out))
  
(def-alien-routine ("cpgline" pgline-raw)  void
  (npt int)
  (xvec (* single-float) :in)
  (yvec (* single-float) :in))

(def-alien-routine ("cpgmove" pgmove-raw) void
  (x float) (y float))

(def-alien-routine ("cpgmtxt" pgmtxt-raw) void
  (side c-string)
  (text c-string)
  (disp float) (coord float) (fjust float))

;; PGNCUR absent
;; PGNUMB absent
;; PGOLIN absent

(def-alien-routine ("cpgopen" pgopen-raw) int
  (device c-string))

(def-alien-routine ("cpgpage" pgpage-raw) void)

(def-alien-routine ("cpgpanl" pgpanl-raw) void
  (ix int) (iy int))

(def-alien-routine ("cpgpap" pgpap-raw) void
  (width float) (aspect float))

(def-alien-routine ("cpgpixl" pgpixl-raw) void
  (idim int) (jdim int) (i1 int) (i2 int) (j1 int) (j2 int)
  (ia (* int) :in)
  (x1 float) (x2 float) (y1 float) (y2 float))


(def-alien-routine ("cpgpoly" pgpoly-raw) void
  (n int)
  (xpts  (* single-float) :in)
  (ypts  (* single-float) :in))
  

(def-alien-routine ("cpgpt" pgpt-raw) void
  (npt int)
  (xvec (* single-float) :in)
  (yvec (* single-float) :in)
  (ptye int))

;; PGPNTS absent
;; PGPT1 absent

(def-alien-routine ("cpgqpos" pgqpos-raw) void
  (x float :out) (y float :out))

(def-alien-routine ("cpgptxt" pgptxt-raw) void
  (x float) (y float) (angle float) (fjust float)
  (text c-string))

;; PGQAH absent
;; PGQCF absent
;; PGQCH absent

(def-alien-routine ("cpgqci" pgqci-raw) void
  (npt int :out))

(def-alien-routine ("cpgqcol" pgqcol-raw) void
  (nlo int :out)
  (nnhi int :out))


(def-alien-routine ("cpgqcir" pgqcir-raw) void
  (icilo int :out)
  (icihi int :out))
;; PGQCLP absent

(def-alien-routine ("cpgqcr" pgqcr-raw) void
  (ci int :in)
  (cr  float :out) (cg  float :out) (cb  float :out))

;; PGQCS absent
;; PGQDT absent
;; PGQFS absent
;; PGQHS absent
;; PGQINF absent
;; PGQITF absent
;; PGQLS absent 
;; PGQLW absent
;; PGQNDT absent 
;; PGQPOS absent
;; PGQTBG absent


(def-alien-routine ("cpgqtxt" pgqtxt-raw) void
  (x float) (y float) (angle float) (fjust float)
  (text c-string)
  (xbox (* single-float) :in)
  (ybox (* single-float) :in))



	      
;; PGQVP absent
  
(def-alien-routine ("cpgqvsz" pgqvsz-raw) void
  (units int)
  (x1 float :out) (x2 float :out)
  (y1 float :out) (y2 float :out))


;; PGQWIN absent

(def-alien-routine ("cpgrect" pgrect-raw) void
  (x1 float) (x2 float) (y1 float) (y2 float))

;; PGRND absent
;; PGRNGE absent

(def-alien-routine ("cpgsah" pgsah-raw) void
  (fs int) (angle float) (barb float))

(def-alien-routine ("cpgsave" pgsave-raw) void)

(def-alien-routine ("cpgunsa" pgunsa-raw) void)

(def-alien-routine ("cpgscf" pgscf-raw) void
  (font int))

(def-alien-routine ("cpgsch" pgsch-raw) void
  (size float))

(def-alien-routine ("cpgsci" pgsci-raw) void
  (ci int))

(def-alien-routine ("cpgscir" pgscir-raw) void
  (icilo int) (icihi int))

(def-alien-routine ("cpgsclp" pgsclp-raw) void
  (state int))

(def-alien-routine ("cpgscr" pgscr-raw) void
  (ci int)
  (cr float) (cg float) (cb float))

(def-alien-routine ("cpgscrl" pgscrl-raw) void
  (dx float) (dy float))

;; PGSCRN absent

(def-alien-routine ("cpgsfs" pgsfs-raw) void
  (fs int))

(def-alien-routine ("cpgshls" pgshls-raw) void
  (ci int)
  (ch float) (cl float) (cs float))

(def-alien-routine ("cpgshs" pgshs-raw) void
  (angle float) (sepn float) (phase float))

(def-alien-routine ("cpgsitf" pgsitf-raw) void
  (itf int))

(def-alien-routine ("cpgslct" pgslct-raw) void
  (id int))

(def-alien-routine ("cpgsls" pgsls-raw) void
  (ls int))

(def-alien-routine ("cpgslw" pgslw-raw) void
  (lw int))

(def-alien-routine ("cpgstbg" pgstbg-raw) void
  (tcbi int))

(def-alien-routine ("cpgsubp" pgsubp-raw) void
  (nxsub int) (nysub int))

(def-alien-routine ("cpgsvp" pgsvp-raw) void
  (xleft float) (xright float) (ybot float) (ytop float))

(def-alien-routine ("cpgswin" pgswin-raw) void
  (x1 float) (x2 float) (y1 float) (y2 float))

;; PGTEXT absent
;; PGTICK absent

(def-alien-routine ("cpgupdt" pgupdt-raw) void)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;; all the following are 'non-standard aliases' and 
;;;;;;;;; will never be present in pgplot-cl
;;
;; PGVECT absent
;; PGVSTD absent
;; PGWEDG absent
;; PGWNAD absent
;; PGADVANCE absent
;; PGBEGIN absent
;; PGCURSE absent
;; PGLABEL absent
;; PGMTEXT absent
;; PGNCURSE absent
;; PGPAPER absent
;; PGPOINT absent
;; PGPTEXT absent
;; PGVPORT absent
;; PGVSIZE absent
;; PGVSTAND absent
;; PGWINDOW absent
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; get vector data address - verbatim from cmucl manual
(defun vector-data-address (vec)
  (let ((base-address (the (unsigned-byte 32)
			(logandc1 7 (kernel:get-lisp-obj-address vec)))))
    (declare (type (unsigned-byte 32) base-address))
    (the (unsigned-byte 32) (+ 8 base-address))))

;; get starting point of vector as system address (sap)
(defun vector-sap (vec)
  (system:int-sap  (vector-data-address vec)))

;; coercion to single float - elaborate form is to prevent underflows
;; that seem to occur on some (Sun..and?) architectures
;; it isn't really clear that this is the right way to do this.
;; -- it would be nice if pgplot supported double floats instead of single
(declaim (inline to-single-float))
(defun to-single-float (x)
  (cond ((< x most-negative-single-float)
	 most-negative-single-float)
	((> x most-positive-single-float)
	 most-positive-single-float)
	;; we use  -normalized because plain least-postitive-float give
	;; underflows on some architectures (dangit..)
	((and (< x least-positive-normalized-single-float)
	      (> x least-negative-normalized-single-float))
	 0.0s0)
	(t
	 (coerce x 'single-float))))


	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; coerce a vector into type (ARRAY SINGLE-FLOAT ...) because
;; that is what pgplot expects 
(defun vector-to-sa-sf (vec)
  (typecase vec
    ((simple-array single-float)
     vec)
    ;;
    (t ;; else duplicate into a new array
     (let ((u (make-array (length vec) :element-type 'single-float)))
       (dotimes (i (length vec))
	 (if (not (numberp (aref vec i)))
	     (error
	      (format
	       t
 "trying to coerce a non-numerical vector into single-float vector : bad element ~A at index ~A"
 (aref vec i) i)))
	 ;; this elaborate coercion to single float prevents
	 ;; over/underflows on some (Sun and ...?) architectures
	 (setf (aref u i) (to-single-float (aref vec i))))
       u))))

;; create a new simple single float array from a 2d array. 
(defun 2d-array-to-sa-sf-copy (a)
  (let* ((ni (array-dimension a 0))
	 (nj (array-dimension a 1))
	 (v  (make-array (* ni nj) :element-type 'single-float))
	 (k 0))
    (declare (type (unsigned-byte 28) ni nj k))
    (dotimes (i  ni)
      (dotimes (j nj)
	(declare (type (unsigned-byte 28) i j))
	(setf (aref v k) (to-single-float (aref a i j)))
	(setf k (+ k 1))))
    v))


;; convert a  2d array into a single float array
(defun 2d-array-to-sa-sf (a)
  (if (typep a '(simple-array single-float (* *)))
      ;; if we can use the array data directly, do so
      (kernel::%array-data-vector a) ;; undocumented cmucl feature
      ;; else use the copying version
      (2d-array-to-sa-sf-copy a)))
      

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; macro to coerce a set of variables to a type
(defmacro ensure-type (type &rest var-list)
  (let ((the-list '()))
    (dolist (var var-list)
      (setq the-list 
	    (append the-list 
		     (list 
		      (list 'setq var (list 'coerce var type))))))
    (cons 'progn the-list)))


(defmacro without-gcing (&body body)
  `(unwind-protect
    (progn (ext:gc-off)
	   ,@body)
    (ext:gc-on)))
    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; now a layer of functions that interfaces between the "raw" 
;; ones and the high level CLOS ones -- often these are identical
;; to the raw ones  (some of these could be macros)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pgarro (x1 y1 x2 y2)
  (ensure-type 'single-float x1 y2 x2 y2)
  (pgarro-raw x1 y1 x2 y2))

(defun pgask  (flag) (pgask-raw flag))

(defun pgbox (xopt xtick nxsub yopt ytick nyxub)
  (ensure-type 'single-float xtick ytick)
  (pgbox-raw xopt xtick nxsub yopt ytick nyxub))

(defun pgcirc (xcent ycent radius)
  (ensure-type 'single-float xcent ycent radius)
  (pgcirc-raw xcent ycent radius))


(defun pgclos () (pgclos-raw))

(defun pgscir (icilo icihi) (pgscir-raw icilo icihi))

(defun pgsclp (state) (pgsclp-raw state))


(defun pgcurs (x y)
  (ensure-type 'single-float x y)
  (multiple-value-bind (dummy x y c) (pgcurs-raw x y)
		       (declare (ignore dummy))
		       (values x y c)))

;; uses the good but slow pgcont algorithm, not pgcons
(defun pgcont (a i1 i2 j1 j2 c n-cont tr)
  (let ((av (2d-array-to-sa-sf a))
	(cv (vector-to-sa-sf c))
	(trv (vector-to-sa-sf tr)))
    (without-gcing
     (let ((a-sap  (vector-sap av))
	   (c-sap  (vector-sap cv))
	   (tr-sap (vector-sap trv)))
       (pgcont-raw a-sap (array-dimension a 1) (array-dimension a 0)
		   ;; indices incremented by 1 for fortran
		   (1+ j1) (1+ j2) (1+ i1) (1+ i2) 
		   c-sap n-cont tr-sap)))))

;; label for 1 contour
(defun pgconl (a i1 i2 j1 j2 c tr label intval minint)
  (let ((av (2d-array-to-sa-sf a))
	(c (float c 1s0))
	(trv (vector-to-sa-sf tr)))
    (without-gcing
     (let ((a-sap  (vector-sap av))
	   (tr-sap (vector-sap trv)))
       (pgconl-raw a-sap (array-dimension a 1) (array-dimension a 0)
		   ;; indices incremented by 1 for fortran
		   (1+ j1) (1+ j2) (1+ i1) (1+ i2) 
		   c tr-sap label intval minint)))))


(defun pgimag (i1 i2 j1 j2 a a1 a2 tr)
  (let ((av (2d-array-to-sa-sf a))
	(trv (vector-to-sa-sf tr)))
    (without-gcing
     (let ((a-sap  (vector-sap av))
	   (tr-sap (vector-sap trv)))
       (pgimag-raw a-sap
		   (array-dimension a 1) (array-dimension a 0)
		   (1+ j1) (1+ j2) (1+ i1) (1+ i2) 
		   a1 a2 tr-sap)))))


(defun pggray (i1 i2 j1 j2 a fg bg tr)
  (let ((av (2d-array-to-sa-sf a))
	(trv (vector-to-sa-sf tr)))
    (without-gcing
     (let ((a-sap  (vector-sap av))
	   (tr-sap (vector-sap trv)))
       (pggray-raw a-sap 
		   (array-dimension a 1) (array-dimension a 0)
		   (1+ j1) (1+ j2) (1+ i1) (1+ i2) 
		   fg bg tr-sap)))))


(defun pgerrb (dir n x y e tt)
  (ensure-type 'single-float tt)
  (let ((xv (vector-to-sa-sf x))
	(yv (vector-to-sa-sf y))
	(ev (vector-to-sa-sf e)))
    (without-gcing
     (let ((x-sap (vector-sap xv))
	   (y-sap (vector-sap yv))
	   (e-sap (vector-sap ev)))
       (pgerrb-raw dir n x-sap y-sap e-sap tt)))))

	


(defun pgenv (xmin xmax ymin ymax just axis)
  (ensure-type 'single-float xmin xmax ymin ymax)
  (pgenv-raw xmin xmax ymin ymax just axis))

(defun pglab (xlabel ylabel toplabel) (pglab-raw xlabel ylabel toplabel))

(defun pglen (units string)
  (multiple-value-bind (dummy xl yl) (pglen-raw units string)
		       (declare (ignore dummy))
		       (values xl yl)))

(defun pgline (npts xvec yvec)
  (let ((xv (vector-to-sa-sf xvec))
	(yv (vector-to-sa-sf yvec))
	(nx (length xvec))
	(ny (length yvec)))
    (if (not (and (<= npts nx) (<= npts ny)))
	(error (format t "pgline: npts=~A nx=~A  ny=~A" npts nx ny)))
    (without-gcing
     (let ((x-sap (vector-sap xv))
	   (y-sap (vector-sap yv)))
       (pgline-raw npts x-sap y-sap)))))

    
(defun pgmove (x y)     (pgmove-raw x y))


(defun pgopen (dev-str) (pgopen-raw dev-str))

(defun pgpap (width aspect)
  (ensure-type 'single-float width aspect)
  (pgpap-raw width aspect))

(defun pgpoly (npts xvec yvec)
  (let ((xv (vector-to-sa-sf xvec))
	(yv (vector-to-sa-sf yvec))
	(nx (length xvec))
	(ny (length yvec)))
    (if (not (and (<= npts nx) (<= npts ny)))
	(error (format t "pgpoly: npts=~A nx=~A  ny=~A" npts nx ny)))
    (without-gcing
     (let ((x-sap (vector-sap xv))
	   (y-sap (vector-sap yv)))
       (pgpoly-raw npts x-sap y-sap)))))


(defun pgpt (npts xvec yvec symbol)
  (let ((xv (vector-to-sa-sf xvec))
	(yv (vector-to-sa-sf yvec))
	(nx (length xvec))
	(ny (length yvec)))
    (if (not (and (<= npts nx) (<= npts ny)))
	(error (format t "pgpt: npts=~A nx=~A  ny=~A" npts nx ny)))
    (without-gcing
     (let ((x-sap (vector-sap xv))
	   (y-sap (vector-sap yv)))
       (pgpt-raw npts x-sap y-sap symbol)))))


(defun pgqcol ()
  (multiple-value-bind (dummy nlo nhi) (pgqcol-raw)
    (declare (ignore dummy))
    (values nlo nhi)))
  

(defun pgrect (x1 x2 y1 y2)
  (ensure-type 'single-float x1 x2 y1 y2)
  (pgrect-raw x1 x2 y1 y2))


(defun pgqtxt (x y angle fjust text xbox ybox)
  (ensure-type 'single-float x y angle fjust)
  (ensure-type 'simple-string text)
  (let ((xv (vector-to-sa-sf xbox))
	(yv (vector-to-sa-sf ybox)))
    (without-gcing
     (let ((x-sap  (vector-sap xv))
	   (y-sap  (vector-sap yv)))
       (pgqtxt-raw x y angle fjust text x-sap y-sap)))))


(defun pgqpos ()
  (multiple-value-bind (dummy x y) (pgqpos-raw)
		       (declare (ignorable dummy))
		       (values x y)))

(defun pgqvsz (units)
  (multiple-value-bind (dummy x1 x2 y1 y2)  (pgqvsz-raw units)
		       (declare (ignorable dummy))
		       (values x1 x2 y1 y2) ))

(defun pgptxt (x y angle fjust text)
  (ensure-type 'single-float x y angle fjust)
  (ensure-type 'simple-string text)
  (pgptxt-raw x y angle fjust text))

(defun pgqcr (ci)  ;; returns (values cr cg cb)
  (multiple-value-bind (dummy cr cg cb) (pgqcr-raw ci)
		       (declare (ignorable dummy))
		       (values cr cg cb)))

(defun pgsah (fs angle barb)  (pgsah-raw fs angle barb))

(defun pgscf (font)   (pgscf-raw font))
(defun pgsch (size)
  (ensure-type 'single-float size)
  (pgsch-raw size))
(defun pgsci (ci) (pgsci-raw ci))
(defun pgsclp (state) (pgsclp-raw state))
(defun pgscr (ci cr cg cb)
  (ensure-type 'single-float cr cg cb)
  (pgscr-raw ci cr cg cb))
(defun pgsfs (fill-style)
  (pgsfs-raw fill-style))
(defun pgshs (angle sepn phase)
  (ensure-type 'single-float angle sepn phase)
  (pgshs-raw angle sepn phase))
(defun pgsitf (itf) (pgsitf-raw itf))
(defun pgsls (ls)     (pgsls-raw ls))
(defun pgslw (lw)     (pgslw-raw lw))
(defun pgstbg (tbci)  (pgstbg-raw tbci))
(defun pgsvp (xleft xright ybot ytop)
  (ensure-type 'single-float xleft xright ybot ytop)
  (pgsvp-raw xleft xright ybot ytop))
(defun pgswin (x1 x2 y1 y2)
  (ensure-type 'single-float x1 x2 y1 y2)
  (pgswin-raw x1 x2 y1 y2))
(defun pgslct (dev) (pgslct-raw dev))
(defun pgupdt ()    (pgupdt-raw))






