;; pgplot-cl version 1.1
;; information in README file 
;; distribution site is http://www.geocities.com/pgplot_cl


(defpackage #:pgplot
  (:use #:common-lisp #:alien #:c-call)
  (:export  #:*pgplot-list*     ;; list of all living devices
	    #:*current-pgplot*  ;; the currently active device
	    #:pgplot-p #:pgplot
	    ;;
	    #:open-device
	    #:close-device
	    #:close-all-devices
	    #:activate-device   ;; not really meant to be used by user
	    ;;
	    ;; macro to set pgplot's qualities temporarily
	    #:with-pgplot-settings
	    ;;
	    ;; getters and setters that work on a device
	    ;; the getters are just :accessors; the setters also set
	    ;; the pgplot library global property if the device being set is
	    ;; the currently active one
	    #:set-character-font    #:get-character-font
	    #:set-character-height  #:get-character-height
	    #:set-text-background-color #:get-text-background-color
	    #:set-color             #:get-color-index
	    #:set-fill-area-style   #:get-fill-area-style
	    #:set-color-index       #:get-color-index
	    #:set-line-style        #:get-line-style
	    #:set-line-width        #:get-line-width
	    #:set-pen-position      #:get-pen-position
	    #:set-arrow-head-style  #:get-arrow-head-style
	    #:set-hatching-style    #:get-hatching-style
	    #:set-clipping-state    #:get-clipping-state
	    #:set-color-index-range #:get-color-index-range
	    #:set-window            #:get-window
	    #:set-viewport          #:get-viewport
;;	    #:set-width
	    #:get-width   
;;	    #:set-aspect-ratio
	    #:get-aspect-ratio
	    #:set-current-pane
	    
	    #:box
	    #:circle
	    #:ellipse
	    #:connect
	    #:contour
	    #:image
	    #:draw-arrow
	    #:erase
	    #:errorbars
	    #:get-cursor-position
	    #:plotlabel  ;; built-in x,y,top labels -- poor positioning
	    #:points
	    #:polygon
	    #:rectangle
	    #:set-colormap-color
	    #:modify-colormap
;;	    #:string-size  ;; rather useless compared to text-bounding-box
	    #:text-bounding-box
	    #:write-text
	    #:world-to-device #:device-to-world ;; coord transformations
	    #:xlabel #:ylabel #:toplabel ;; better labels than plotlabel

	    #:hist
           ))


(in-package pgplot)

(eval-when (load eval compile)
	   #+cmu (progn
		   ;; need to load pgplot-cmucl-libs before pgplot-cmucl
		   ;; to prevent barfing on unrecongnized foreign symbols
		   ;; inside pgplot-cmucl
		   (when (not (ignore-errors (ext:search-list "pgplot:")))
		     (error 
		      "You seem not to have set (ext:search-list \"pgplot\""))
		   ;;
		   (load "pgplot:pgplot-cmucl-libs") ;; libraries
		   (load "pgplot:pgplot-cmucl"))     ;; and cmucl ffi code
	   ;;
	   ;; ADD MORE LISP IMPLEMENTATIONS HERE 
	   ;;
	   #-cmu (error "only cmucl supported by pgplot (so far)")
	   )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; define class for a pgplot plot, with attributes included --
;; thus the attributes get carried around by a plot, and changing
;; one plot's attributes doesn't mess up other plots
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defclass pgplot ()
  ((name :accessor get-name :initform "untitled" :initarg :name)
   ;; numerical device id of plot
   (id   :accessor get-id :initarg :id)
   ;; is the device open?
   (is-open :accessor is-open :initform nil :initarg :is-open)
   ;; symbolic representation of device
   (device :accessor get-device :initform nil :initarg :device) 
   ;;
   (character-font    :accessor get-character-font)
   (character-height  :accessor get-character-height)
   (color-index       :accessor get-color-index)
   (fill-area-style   :accessor get-fill-area-style)
   (line-style        :accessor get-line-style)
   (line-width        :accessor get-line-width)
   (pen-position      :accessor get-pen-position)
   (arrow-head-style  :accessor get-arrow-head-style)
   (hatching-style    :accessor get-hatching-style)
   (clipping-state    :accessor get-clipping-state)
   (text-background-color-index
                      :accessor get-text-background-color-index)
   ;;
   (viewport          :accessor get-viewport) ;; '(XLEFT XRIGHT YBOT YTOP)
   (window            :accessor get-window)   ;; world coords '(x1 x2 y1 y2)
   (aspect-ratio      :accessor get-aspect-ratio)
   (width             :accessor get-width)
   ;;
   (color-index-range :accessor get-color-index-range)
   ;;
   ;; we use colormaps, rather than individual color indices
   (colormap          :accessor get-colormap)
   (default-colormap  :accessor get-default-colormap)
   ;; last cursor position, or NULL if none
   (last-curs-x :accessor get-last-curs-x :initform nil)
   (last-curs-y :accessor get-last-curs-y :initform nil)
   ))


(defun pgplot-p (x)
  (eq (type-of x) 'pgplot))

;; a colormap is three vectors of red,blue,green
;; determining color indices nos. 0..npts-1
;; we keep the first 16 at the default values
(defstruct colormap
  (name nil) ;; a general name for this colormap
  npoints    ;; how many color points?
  (color-alist nil) ;; an alist for the keyword-named colors in this map
  ;; single-float vectors for the actual colors
  (red-vec  nil :type (or null (simple-array single-float (*))))
  (green-vec nil :type (or null (simple-array single-float (*))))
  (blue-vec nil :type (or null (simple-array single-float (*))))
  ;; a list of the form eg '((:default 34 76) (:gray 34 76) (:rainbow 18 77) ...)
  ;; that delimits inclusive color ranges in the map, for making
  ;; images - each colormap should have a default range
  (named-ranges (list (list :default 0 16))))


;; name to color index mapping for default device - background/default
;; are black/white or white/black depending on device 
(defvar *default-colormap-color-alist*
  '((:BACKGROUND . 0) (:DEFAULT . 1) (:RED . 2) (:GREEN . 3) (:BLUE . 4)
    (:CYAN . 5) (:MAGENTA . 6) (:YELLOW . 7) (:ORANGE . 8)
    (:GREEN+YELLOW . 9) (:GREEN+CYAN . 10) (:BLUE+CYAN . 11)
    (:BLUE+MAGENTA . 12) (:RED+MAGENTA . 13)
    (:DARK-GREY . 14) (:LIGHT-GREY . 15) (:DARK-GRAY . 14) (:LIGHT-GRAY . 15)))

;; generate the default colormap for current device - 16<=nelem<=255 --
;; nelem>16 only if rest of colormap is to be filled in with other colors
;; fixme - currently uses 100 colors, which is OK for X11, but could
;; adjust automatically to number of available colors.  also, our
;; color colormap is not so good - eg, no yellow
(defun make-default-colormap ()
"make a default colormap with the normal colors, a :grey
component, a :color rainbow component, and a :heat component"
  (let* ((nelem 100) ;; ok for X
	 (rv  (make-array nelem :element-type 'single-float))
	 (gv  (make-array nelem :element-type 'single-float))
	 (bv  (make-array nelem :element-type 'single-float))	      
	 (cm (make-colormap
	      :name :default
	      :npoints nelem :color-alist *default-colormap-color-alist*
	      :red-vec rv :green-vec gv :blue-vec bv
	      )))
    (loop for i from 0 below 16
	  do
	  (multiple-value-bind (r g b) (pgqcr i)
	    (setf (aref rv i) r)
	    (setf (aref gv i) g)
	    (setf (aref bv i) b)))
    ;; make the next highest ones be gray scale
    (insert-gray-colormap-component cm 16 31 '(:gray :grey))
    ;; and then a rainbow
    (insert-rainbow-colormap-component cm 32 76 '(:color :colour))
    ;; and then a heat map
    (insert-heat-colormap-component cm 77 99 :heat)
    cm ;; return value
    ))

;; insert a colormap component into indices - func is a function of
;; (i-imin)/(imax-imin) that returns the red,green,blue values for
;; this colormap, over a domain [0,1].  name is the keyword name of
;; this colormap - delete-overlapping-ranges tells us to
;; remove any other ranges that were overwritten by this insertiion
(defun insert-colormap-component (cm imin imax name func &key
				     (delete-overlapping-ranges t))
  (declare (type colormap cm)
	   (type (unsigned-byte 16) imin imax)
	   (type (function (float) (values float float float)) func))
  (when (or (>= imin imax) (>= imax (colormap-npoints cm)))
    (error "bogus indices imin=~a imax=~a" imin imax))
  (loop
   with dx = (float (- imax imin))
   with rv = (colormap-red-vec cm)
   with gv = (colormap-green-vec cm)
   with bv = (colormap-blue-vec cm)
   for i of-type (unsigned-byte 16) from imin to imax
   for j of-type (unsigned-byte 16) from 0 to (- imax imin)
   for x = (/ (float (- i imin)) dx)
   do (multiple-value-bind (r g b) (funcall func x)
	(setf (aref rv i) (coerce r 'single-float)
	      (aref gv i) (coerce g 'single-float)
	      (aref bv i) (coerce b 'single-float))))
  ;;
  (when delete-overlapping-ranges
    (setf (colormap-named-ranges cm)
	  (remove-if (lambda (range)
		       (not (or (>= (second range) imax)
				(<= (third range) imin))))
		     (colormap-named-ranges cm))))
  ;;
  (cond ((listp name)
	 (dolist (one-name name)
	   (push (list one-name imin imax)  (colormap-named-ranges cm))))
	(t
	 (push (list name imin imax)  (colormap-named-ranges cm)))))

;; a rainbow colormap made up of 3 gaussians - widths are
;; empirical (ie, obtained by fiddling)
(defun insert-rainbow-colormap-component (cm imin imax name)
  (labels ((gauss (x x0 sigma)
		  (declare (type single-float x x0 sigma))
		  (exp (- (* 0.5 (expt (/ (- x x0) sigma) 2)))))
	   (func (x)
		 (declare (type float x))
		 (values (gauss x 0.0 0.40)
			 (gauss x 0.5 0.33)
			 (gauss x 1.0 0.40))))
    (insert-colormap-component cm imin imax name #'func)))

;; linearly rising grayscale map
(defun insert-gray-colormap-component (cm imin imax name)
  (flet ((func (x)
	       (declare (type float x))
	       (values x x x)))
    (insert-colormap-component cm imin imax name #'func)))

;; 'heat' colomap rising from black to red to yellow to white,
;;  made up of 3 successive ramps of red,green,blue
(defun insert-heat-colormap-component (cm imin imax name)
  (flet ((func (x)
	       (declare (type float x))
	       (values
		(min 1.0 (max 0.0 (* 3.0 x)))
		(min 1.0 (max 0.0 (* 3.0 (- x 0.3))))
		(min 1.0 (max 0.0 (* 3.0 (- x 0.6)))))))
    (insert-colormap-component cm imin imax name #'func)))


(defmethod modify-colormap ((p pgplot) imin imax colormap-name &key func)
  (declare (type (unsigned-byte 16) imin imax)
	   (type symbol colormap-name)
	   (type (or null (function (float) (values float float float))) func))
  (when (not func)
    (cond ((or (eq colormap-name :gray) (eq colormap-name :grey))
	   (insert-gray-colormap-component (get-colormap p) imin imax
					   (list :gray :grey)))
	  ((or (eq colormap-name :color) (eq colormap-name :colour))
	   (insert-gray-colormap-component (get-colormap p) imin imax
					   (list :color :colour)))
	  ((eq colormap-name :heat)
	   (insert-heat-colormap-component (get-colormap p) imin imax :heat))
	  (t (error "acceptable colormap-names are :color :colour :gray :grey :heat"))))
  (when func
    (when (not colormap-name)
      (error "You must name the colormap range using :colormap-name :your-chosen-name"))
    (insert-colormap-component (get-colormap p) imin imax colormap-name func))
  (activate-device p))

	
	 
  
		     

		     

;; set color map in active device to colormap cm
(defun set-colormap (cm)
  (dotimes (i (colormap-npoints cm))
    (declare (type (unsigned-byte 16) i))
    (pgscr i
	   (aref (colormap-red-vec   cm)  i)
	   (aref (colormap-green-vec cm)  i)
	   (aref (colormap-blue-vec  cm)  i))))
	   

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; now the higher level (user) versions of the pgplot routines
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *current-pgplot* nil) ;; the last plot used
(defvar *pgplot-list* nil) ;; list of all living plots

;; how does one get this directly from pgplot?
(defconstant +max-pgplot-open-devices+ 8)


(defvar *pgplot-devices-alist*
;;   name      str        need-filename?  nil, t, or :maybe
  '((:XWINDOW  "/xwindow" nil)  
    (:X        "/xwindow" nil)
    (:X11      "/xwindow" nil)
    (:NULL     "/null"    nil)
    (:GIF      "/gif"     t)
    (:VGIF     "/vgif"    t)
    (:PSLAND   "/cps"     t)          ;; we use only color postscript
    (:PSPORT   "/vcps"    t)  
    (:PS       "/cps"     t)))        ;; square postscript
    

(defun open-device (device &key (plot-name "untitled") (filename nil)
			        (square nil) (invert nil))
  "open a pgplot device, where DEVICE is one of 
:X11 :NULL :GIF :VGIF :PSLAND :PSPORT :PS.
PLOT-NAME is an optional name; FILENAME is the file to which to
write if device type chosen targets file; SQUARE is a flag
that tries to make the final plot square; INVERT inverts normal color
scheme"
  (if (>= (length *pgplot-list*) +max-pgplot-open-devices+)
      (error "pgplot supports only ~A simultaneously open devices.
   You must close one or more devices before making more plots
   [eg, use (pgplot:close-all-devices) ] "
	     +max-pgplot-open-devices+))
  (let* ((dev-list (assoc device *pgplot-devices-alist*))
	 (dev-str  (second dev-list))
	 (needs-file (third dev-list))
	 (dev-id 0)
	 (p nil))
    ;;
    (if (not dev-str)
	(error "No device ~A -- Available pgplot devices are ~A"
	       device (map 'list #'car  *pgplot-devices-alist*)))
    (if (and needs-file (not filename))
	(error "Need to specify :FILENAME for device ~A" device))
    (if (and (not needs-file) filename)
	(error "Device ~A does not take a :FILENAME" device))
    ;;
    ;;
    (if (eq device :ps) (setf square t)) ;; special square device
    (setf dev-id (pgopen-raw (concatenate 'string filename dev-str)))
    ;;
    (setf p (make-instance 'pgplot :id dev-id :device device :name plot-name))
    ;; mark device as open
    (setf (is-open p) t)
    ;;
    ;; and set defaults in p, but don't call pgplot routines yet
    (setf (get-character-font p)   2)    ;; ROMAN
    (setf (get-character-height p) 1.7)  ;; biggish
    (setf (get-color-index p) 1)
    (setf (get-fill-area-style p) 2)     ;; OUTLINE
    (setf (get-line-style p) 1)          ;; full line
    (setf (get-line-width p) 2)          ;; double minimal thickness
    (setf (get-pen-position p) '(0.0 0.0)) ;; x,y
    (setf (get-arrow-head-style p) '(2 45.0 0.4)) ;; FS ANGLE BARB
    (setf (get-hatching-style p) '(45.0 2.0 0.0)) ;; ANGLE SEPN PHASE
    (setf (get-clipping-state p) 1)      ;; CLIPPING ENABLED
    (setf (get-text-background-color-index p) -1)
    (setf (get-viewport p) '(0.20 0.95 0.15 0.90)) ;; what looks good
    (setf (get-window p)   '(0.0 1.0 0.0 1.0))
    (setf (get-default-colormap p) (make-default-colormap))
    (set-colormap (get-default-colormap p))
    (setf (get-colormap p) (get-default-colormap p))
    (setf (get-color-index-range p)
	  (list 1 (colormap-npoints (get-colormap p))))

    ;;
    (push p *pgplot-list*) ;; add to global list
    (pgask 0);; always have this - having to hit RET is annoying
    ;; if :SQUARE, set the aspect ratio to 1
    (if square  (pgpap 0.0 1.0))
    ;; set aspect ratio
    (multiple-value-bind (x1 x2 y1 y2) (pgqvsz 1)
      (if (or (= x1 1) (= y1 1)) (print "Yow!")) ;; avoid unused var warning
      (setf (get-width  p)  (* 25.4 x2)) ;; inches to mm
      (setf (get-aspect-ratio p)  (/ y2 x2)))
    
    (activate-device p) ;; make current active device match this open dev
    (if invert
	(progn
	  (set-colormap-color p 0 1 1 1)
	  (set-colormap-color p 1 0 0 0)))
	
    ;; eliminate the worst colormap flashing for x devices by erasing
    ;; first
    (if (member device '(:XWINDOW :X :X11))
	(erase p))
    ;;
    p))


;; close device, ensuring that current device remains set
(defmethod close-device ((p pgplot))
"close a pgplot device object"
  (if (is-open p)  ;; ignore closed devices
      (progn
	(setf (is-open p) nil) ;; mark device as closed
	(let ((old-dev (if *current-pgplot* (get-id *current-pgplot*) nil))
	      (this-dev (get-id p)))
	  (pgslct this-dev)
	  (pgask 0) ;; disable the silly return key requirement
	  (pgupdt)
	  (pgclos)
	  ;; restore the old device, if it was not the new device
	  (if (and old-dev (not (= this-dev old-dev)))
	      (pgslct old-dev))
	  (setf *pgplot-list* (remove p *pgplot-list*))
	  t
	  ))
    nil))

(defun close-all-devices ()
"close all open devices contained in pgplot::*pgplot-list*"
  (dolist (p *pgplot-list*) (close-device p))
  (setf *pgplot-list* nil)) ;; just in case a closed dev was on list


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; activate a device p -- ie, if p is not the active device in
;; *current-pgplot*, make p be the active dev. and set up the
;; active dev's parameters to be p's
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmethod activate-device ((p pgplot))
"make pgplot device P the active one"
  (cond ((not (is-open p))
	 (error "Tried to activate a closed pgplot device"))
	((eq p *current-pgplot*)
	 t) ;; already open
	(t ;; else restore settings
	 (setf *current-pgplot* p)
	 (pgslct   (get-id p))
	 (pgscf    (get-character-font p))
	 (pgsch    (get-character-height p))
	 (pgstbg   (get-text-background-color-index p))
	 (pgsci    (get-color-index p))
	 (pgsfs    (get-fill-area-style p))
	 (pgsls    (get-line-style p))
	 (pgslw    (get-line-width p))
	 (apply #'pgmove (get-pen-position p))
	 (apply #'pgsah  (get-arrow-head-style p))
	 (apply #'pgshs  (get-hatching-style p))
	 (apply #'pgsvp  (get-viewport p))
	 (apply #'pgswin (get-window p))
	 (pgsclp   (get-clipping-state p))
	 (set-colormap (get-colormap p))
	 (apply #'pgscir   (get-color-index-range p))
	 t)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; at the end of routines that move the pen, we call
;; fix-pen-position to update our remembered pen position
;; for this device (fix-pen-position is internal to this package)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun fix-pen-position (p)
  (multiple-value-bind (x y) (pgqpos)
			(setf (get-pen-position p) (list x y))))
			

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar *character-font-alist* '((:DEFAULT . 1) (:ROMAN . 2) (:ITALIC . 3)
				 (:SCRIPT . 4)))

(defgeneric set-character-font ((p pgplot) font)
  (:documentation
   "set the current character font of device P, where font
is either an integer 1-4, or one of :DEFAULT :ROMAN :ITALIC :SCRIPT"))

(defmethod set-character-font ((p pgplot) (font integer))
  (if (or (< font 1) (> font 4))
      (error "Bad font number ~A" font))
  (setf (get-character-font p) font)
  (if (eq p *current-pgplot*)
      (pgscf font))
  t)
;;
(defmethod set-character-font ((p pgplot) (font symbol))
  (let ((nf (cdr (assoc font *character-font-alist*))))
    (if (not nf)
	(error
	 "Invalid font ~A -- Available fonts are ~A"
	 font (map 'list #'car  *character-font-alist*)))
    (set-character-font p nf)))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod set-character-height ((p pgplot) (height number))
    "set the character height of pgplot device P to a number.
Default character height is 2.0"
  (if (or (< height 0.0) (> height 50))
      (error "Bad font number ~A -- want 0<height<50" height))
  (setf (get-character-height p) height)
  (if (eq p *current-pgplot*)
      (pgsch height))
  t)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defgeneric set-text-background-color ((p pgplot) c)
  (:documentation
   "set the text background color of pgplot device P to
C, where C is either an integer from the colormap, or a
symbol for named colors"))

(defmethod set-text-background-color ((p pgplot) (ci integer))
  (setf (get-text-background-color-index p) ci)
  (if (eq p *current-pgplot*)
      (pgstbg ci))
  t)
;;
(defmethod set-text-background-color ((p pgplot) (cs symbol))
  (let* ((c-alist (cons '(:TRANSPARENT . -1) ;; add special transparent color
			(colormap-color-alist (get-colormap p))))
	 (ci (cdr (assoc cs  c-alist))))
    (if (not ci)
	(error "Color ~A not available.  Available colors are: ~A"
	       cs (map 'list #'car c-alist)))
    (set-text-background-color p ci)))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;    

(defgeneric set-color ((p pgplot) c)
  (:documentation
   "set the plotting color of pgplot device P to
C, where C is either an integer from the colormap, or a
symbol for named colors"))

(defmethod set-color ((p pgplot) (ci integer))
  (setf (get-color-index p) ci)
  (if (eq p *current-pgplot*)
      (pgsci ci))
  t)

(defmethod set-color ((p pgplot) (cs symbol))
  (let ((ci (cdr (assoc cs (colormap-color-alist (get-colormap p))))))
    (if (not ci)
	(error "Color ~A not available.  Available colors are: ~A"
	       cs (map 'list #'car
		       (colormap-color-alist (get-colormap p)))))
    (set-color p ci)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar *fill-area-style-alist* '((:SOLID . 1) (:OUTLINE . 2)
				  (:HATCHED . 3) (:CROSS-HATCHED . 4)))

(defgeneric set-fill-area-style ((p pgplot) s)
  (:documentation
   "set the fill area style of pgplot device P to
S, where S is either an integer or one of
:SOLID :OUTLINE :HATCHED :CROSS-HATCHED"))

(defmethod set-fill-area-style ((p pgplot) (style integer))
  (if (or (< style 0) (> style 4))
      (error "out of range fill area style ~A" style))
  (setf (get-fill-area-style p) style)
  (if (eq p *current-pgplot*)
      (pgsfs style))
  t)

(defmethod set-fill-area-style ((p pgplot) (style symbol))
  (let ((ns (cdr (assoc style *fill-area-style-alist*))))
    (if (not ns)
	(error "Invalid fill-area style ~A -- Available styles are ~A"
	       style (map 'list #'car  *fill-area-style-alist*)))
    (set-fill-area-style p ns)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar *linestyle-alist* '((:SOLID . 1) (:DASHED . 2) (:DASH-DOT . 3)
			    (:DOTTED . 4) (:DASH-DOT-DOT-DOT . 5)))

(defgeneric set-line-style ((p pgplot) s)
  (:documentation
   "set the line style of pgplot device P to
S, where S is either an integer or one of
:SOLID :DASHED :DASHED-DOT :DOTTED :DASH-DOT-DOT-DOT"))
;;
(defmethod set-line-style ((p pgplot) (ls integer))
  (if (or (< ls 1) (> ls 5))
      (error "out of range line style ~A" ls))
  (setf (get-line-style p) ls)
  (if (eq p *current-pgplot*)
      (pgsls ls))
  t)
;;
(defmethod set-line-style ((p pgplot) (lss symbol))
  (let ((ls (cdr (assoc lss *linestyle-alist*))))
    (if (not ls)
	(error "Invalid line style ~A -- Available styles are ~A"
	       lss (map 'list #'car  *linestyle-alist*)))
    (set-line-style p ls)))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric set-line-width ((p pgplot) lw)
  (:documentation
   "set the line width pgplot device P to
integer value LW"))

(defmethod set-line-width ((p pgplot) (lw integer))
  (if (or (< lw 1) (> lw 1000))
      (error "out of range line width ~A -- expect 1<=width<=1000" lw))
  (setf (get-line-width p) lw)
  (if (eq p *current-pgplot*)
      (pgslw lw))
  t)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric set-arrow-head-style ((p pgplot) fs angle barb)
  (:documentation
   "set the arrow head style of pgplot.  FS can be 1, 2, :FILLED,
:OUTLINE.  ANGLE is the angle of the tip; BARB is the fraction of the
triangular arrow-head that is cut away from the back."))

(defmethod set-arrow-head-style
  ((p pgplot) fs (angle real) (barb real))
  (cond ((eq fs :FILLED) (setq fs 1))
	((eq fs :OUTLINE) (setq fs 2))
	((and (integerp fs) (or (= fs 1) (= fs 2))) t)
	(t
	 (error
	  "Invalid arrow head style ~A. Available styles are ~A"
	  fs '(:FILLED :OUTLINE 1 2))))
  (setf angle (float angle 1.0))
  (setf barb (float barb 1.0))
  (setf (get-arrow-head-style p) (list fs angle barb))
  (if (eq p *current-pgplot*)
      (pgsah fs angle barb))
  t)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod set-hatching-style
    ((p pgplot) (angle real) (sepn real) (phase real))
    "set hatching style of pgplot device P.
ANGLE is angle wrt horizontal.  SEPN is separation in
units of the view surface.  PHASE in [0,1] is the offset
within the separation"
  (setf (get-hatching-style p) (list angle sepn phase))
  (if (eq p *current-pgplot*)
      (pgshs angle sepn phase))
  t)

(defmethod set-clipping-state ((p pgplot) (state integer))
  "set clipping state of pgplot device P.  STATE=1 means clip at
edge of viewport, 0 means don't clip"
  (cond ((or (= state 0) (= state 1))  t) ;; OK
	(t (error "Bad state in set-clipping-state")))
  (if (or (eq state t) (= state 1)) (setq state 1) (setq state 0))
  (setf (get-clipping-state p) state)
  (if (eq p *current-pgplot*)
      (pgsclp state))
  t)


(defmacro with-pgplot-settings
    ((pgplot &key 
	(font nil)
	(character-height nil)
	(text-background-color nil)
	(color nil)
	(fill-area-style nil)
	(line-style nil)
	(line-width nil)
	(arrow-head-style nil);; list (fs angle barb)
	(hatching-style nil);; list (angle sepn phase)
	(clipping-state nil))
     &body body)
  "macro to run BODY with pgplot's  properties temporarily
set to give values.
 eg draw a blue box
  (with-pgplot-settings (pgplot :color :blue)
       (pgplot:box pgplot))
available settings are 
 :FONT :CHARACTER-HEIGHT :TEXT-BACKGROUND-COLOR :COLOR
 :FILL-AREA-STYLE :LINE-STYLE :CLIPPING-STATE
 :ARROW-HEAD-STYLE  (a list of fs,angle,barb)
 :HATCHING-STYLE    (a list of angle,sepn,phase)
"
  (let ((p (gensym "pgplot"))
	(sym-old-character-font
	 (when font (gensym "old-character-font")))
	(sym-old-character-height
	 (when character-height (gensym "old-character-height")))
	(sym-old-text-background-color
	 (when text-background-color (gensym "old-text-background-color")))
	(sym-old-color
	 (when color (gensym "old-color")))
	(sym-old-fill-area-style
	 (when fill-area-style (gensym "old-fill-area-style")))
	(sym-old-line-style
	 (when line-style (gensym "old-line-style")))
	(sym-old-line-width
	 (when line-width (gensym "old-line-width")))
	(sym-old-arrow-head-style
	 (when arrow-head-style (gensym "old-arrow-head-style")))
	(sym-old-hatching-style
	 (when hatching-style (gensym "old-hatching-style")))
	(sym-old-clipping-state
	 (when clipping-state (gensym "old-clipping-state")))
	;;
	(sym-font
	 (when font (gensym "character-font")))
	(sym-character-height
	 (when character-height (gensym "character-height")))
	(sym-text-background-color
	 (when text-background-color (gensym "text-background-color")))
	(sym-color
	 (when color (gensym "color")))
	(sym-fill-area-style
	 (when fill-area-style (gensym "fill-area-style")))
	(sym-line-style
	 (when line-style (gensym "line-style")))
	(sym-line-width
	 (when line-width (gensym "line-width")))
	(sym-arrow-head-style
	 (when arrow-head-style (gensym "arrow-head-style")))
	(sym-hatching-style
	 (when hatching-style (gensym "hatching-style")))
	(sym-clipping-state
	 (when clipping-state (gensym "clipping-state"))))
  `(let*
    ,(append
      `((,p ,pgplot)) ;; bind pgplot to new var in case it is a complex form
      ;;
      ;; sym-font is variable evaluated version of FONT, to
      ;; prevent multiple-evaluation of FONT form.
      ;; sym-old-character-font is a variable for saving old version
      (when font
	`((,sym-font ,font)
	  (,sym-old-character-font (when ,sym-font (get-character-font ,p)))))
      (when character-height
	`((,sym-character-height ,character-height)
	  (,sym-old-character-height
	   (when ,sym-character-height (get-character-height ,p)))))
      (when text-background-color
	`((,sym-text-background-color ,text-background-color)
	  (,sym-old-text-background-color
	   (when ,sym-text-background-color
	     (get-text-background-color-index ,p)))))
      (when color
	`((,sym-color ,color)
	  (,sym-old-color (when ,sym-color (get-color-index ,p)))))
      (when fill-area-style
	`((,sym-fill-area-style ,fill-area-style)
	  (,sym-old-fill-area-style
	   (when ,sym-fill-area-style (get-fill-area-style ,p)))))
      (when line-style
	`((,sym-line-style ,line-style)
	  (,sym-old-line-style
	   (when ,sym-line-style (get-line-style ,p)))))
      (when line-width
	`((,sym-line-width ,line-width)
	  (,sym-old-line-width
	   (when ,sym-line-width (get-line-width ,p)))))
      (when arrow-head-style
	`((,sym-arrow-head-style ,arrow-head-style)
	  (,sym-old-arrow-head-style
	   (when ,sym-arrow-head-style (get-arrow-head-style ,p)))))
      (when hatching-style
	`((,sym-hatching-style ,hatching-style)
	  (when ,sym-hatching-style (hatching-style (get-hatching-style ,p)))))
      (when clipping-state
	`((,sym-clipping-state ,clipping-state)
	  (clipping-state (when ,sym-clipping-state (get-clipping-state ,p)))))
      )
    (unwind-protect
	 (progn
	   ,@(append
	     (when font
	       `((when ,sym-font (set-character-font ,p ,sym-font))))
	     (when character-height
	       `((when ,sym-character-height
		   (set-character-height ,p ,sym-character-height))))
	     (when text-background-color
	       `((when ,sym-text-background-color
		   (set-text-background-color ,p ,sym-text-background-color))))
	     (when color
	       `((when ,sym-color (set-color ,p ,sym-color))))
	     (when fill-area-style 
	       `((when ,sym-fill-area-style
		   (set-fill-area-style ,p ,sym-fill-area-style))))
	     (when line-style
	       `((when ,sym-line-style (set-line-style ,p ,sym-line-style))))
	     (when line-width
	       `((when ,sym-line-width (set-line-width ,p ,sym-line-width))))
	     (when arrow-head-style ;; this one takes several args in list
	       `((when ,sym-arrow-head-style
		   (apply set-arrow-head-style ,p ,sym-arrow-head-style))))
	     (when hatching-style ;; this one takes several args in list
	       `((when ,sym-hatching-style
		   (apply set-hatching-style ,p ,sym-hatching-style))))
	     (when clipping-state
		  `((when ,sym-clipping-state
		      (set-clipping-state ,p ,sym-clipping-state)))))
	   ;; now run the body
	   ,@body
	   ;; and restore the old values
	   (progn
	     ,@(append
	       (when font
		 `((when ,font
		    (set-character-font ,p ,sym-old-character-font))))
	       (when character-height
		 `((when ,character-height
		     (set-character-height ,p ,sym-old-character-height))))
	       (when text-background-color
		 `((when ,text-background-color
		     (set-text-background-color
		      ,p ,sym-old-text-background-color))))
	       (when color
		 `((when ,color (set-color ,p ,sym-old-color))))
	       (when fill-area-style 
		 `((when ,fill-area-style
		     (set-fill-area-style ,p ,sym-old-fill-area-style))))
	       (when line-style
		 `((when ,line-style (set-line-style ,p ,sym-old-line-style))))
	       (when line-width
		 `((when ,line-width (set-line-width ,p ,sym-old-line-width))))
	       (when arrow-head-style ;; this one takes several args in list
		 `((when ,arrow-head-style
		     (apply set-arrow-head-style ,p ,sym-old-arrow-head-style))))
	       (when hatching-style ;; this one takes several args in list
		 `((when ,hatching-style
		     (apply set-hatching-style ,p ,sym-old-hatching-style))))
	       (when clipping-state
		 `((when ,clipping-state
		     (set-clipping-state ,p ,sym-old-clipping-state))))
	       )))))))
    
			       
				       




(defmethod set-color-index-range ((p pgplot) (ilo integer) (ihi integer))
  "set the color index rangle of pgplot device P to be ILO..IHI.
This is used for drawing imagemaps or greymaps"
  (setf (get-color-index-range p) (list ilo ihi))
  (if (eq p *current-pgplot*)
      (pgscir ilo ihi))
  t)

(defmethod set-viewport ((p pgplot) (xleft real) (xright real)
			 (ybot real) (ytop real))
  "set the viewport of device P.  This is the region in absolute
coordinates over which plotting is active.  XLEFT, XRIGHT, YBOT, YTOP
are all in [0,1]"
  (setf (get-viewport p) (list xleft xright ybot ytop))
  (if (eq p *current-pgplot*)
      (pgsvp xleft xright ybot ytop))
  t)


(defmethod set-window ((p pgplot) (x1 real) (x2 real)
		       (y1 real) (y2 real))
  "Set the window (user coordinates) of pgplot device P within the
current viewport.  X1,X2,Y1,Y2 are the user coordinates to map
to the viewport."
  (ensure-type 'single-float x1 x2 y1 y2)
  (setf (get-window p) (list x1 x2 y1 y2))
  (if (eq p *current-pgplot*)
      (pgswin x1 x2 y1 y2))
  t)


;;
(defmethod move-to ((p pgplot) (x real) (y real)) 
  (setf (get-pen-position p) (list x y))
  (if (eq p *current-pgplot*)
      (pgmove x y))
  t)  

(defgeneric box
  ((p pgplot) &key
   (size nil) (font nil) (color nil)
   (x-axis nil) (y-axis nil)
   (x-bottom t) (x-top t)
   (y-left t) (y-right t)
   (x-grid nil) (y-grid nil)
   (x-invert-ticks nil) (y-invert-ticks nil)
   (x-log nil) (y-log nil)
   (x-num-labels-bottom t)   (y-num-labels-left t)
   (x-num-labels-top nil)    (y-num-labels-right  nil)
   (x-project-ticks nil) (y-project-ticks nil)
   (x-major-ticks t) (y-major-ticks t)
   (x-minor-ticks t) (y-minor-ticks t)
   (y-vertical-num-labels t)
   (x-force-decimal nil) (y-force-decimal nil)
   (x-force-exponential nil) (y-force-exponential nil)
   (x-tick 0.0) (y-tick 0.0) ;; automatic by default
   (nxsub 0) (nysub 0))
  (:documentation
   " Draw a box for pgplot device P.
Keywords are:
  SIZE             -  size used for labels (real)
  FONT             -  which font to use (number or symbol)
  X-AXIS           -  draw central x axis?   (default NIL)
  Y-AXIS           -  draw central y axis?   (default NIL)
  X-BOTTOM, X-TOP  -  draw x axes            (default T)
  Y-LEFT, Y-RIGHT  -  draw y axis            (default T)
  X-GRID, Y-GRID   -  draw grids in X,Y      (default NIL)
  X-INVERT-TICKS, Y-INVERT-TICKS - ticks project outward (default NIL)
  X-LOG, Y-LOG     -  log axes               (default NIL)
  X-NUM-LABELS-BOTTOM, Y-NUM-LABELS-LEFT - numerical axes labels (default NIL)
  X-NUM-LABELS-TOP, Y-NUM-LABELS-RIGHT   - numerical axes labels (default T)
  X-PROJECT-TICKS, Y-PROJECT-TICKS - project ticks outward, ignored if
                                     X-INVERT-TICKS set (default NIL)
  X-MAJOR-TICKS, Y-MAJOR-TICKS  - draw big interval ticks   (default T)
  X-MINOR-TICKS, Y-MINOR-TICKS  - draw small interval ticks (default T)
  Y-VERTICAL-NUM-LABELS  - make y axes labels normal vertical text (default T)
  X-FORCE-DECIMAL, Y-FORCE-DECIMAL - force axis label to be of form 1.234
                                     (default NIL)
  X-FORCE-EXPONENTIAL, Y-FORCE-EXPONENTIAL - force exponential notation for
                                             axis labels (default NIL)
  X-TICK, Y-TICK   -  world coordinate interval between major tick marks,
                      0 means choose automatically (default 0)
  NXSUB, NYSUB     -  the number of subintervals to divide the major
                      coordinate interval into, 0 means choose automatically
                      (default 0)
"))
  
(defmethod box ((p pgplot) &key
		(size nil) (font nil) (color nil)
		(x-axis nil) (y-axis nil)
		(x-bottom t) (x-top t)
		(y-left t) (y-right t)
		(x-grid nil) (y-grid nil)
		(x-invert-ticks nil) (y-invert-ticks nil)
		(x-log nil) (y-log nil)
		(x-num-labels-bottom t)   (y-num-labels-left t)
		(x-num-labels-top nil)    (y-num-labels-right  nil)
		(x-project-ticks nil) (y-project-ticks nil)
		(x-major-ticks t) (y-major-ticks t)
		(x-minor-ticks t) (y-minor-ticks t)
		(y-vertical-num-labels t)
		(x-force-decimal nil) (y-force-decimal nil)
		(x-force-exponential nil) (y-force-exponential nil)
		(x-tick 0.0) (y-tick 0.0) ;; automatic by default
		(nxsub 0) (nysub 0))
  ;;
  (activate-device p)
  (let ((xopt (concatenate 'string
			   (if x-axis "A" "")
			   (if x-bottom "B" "")
			   (if x-top "C" "")
			   (if x-grid "G" "")
			   (if x-invert-ticks "I" "")
			   (if x-log "L" "")
			   (if x-num-labels-bottom "N" "")
			   (if x-num-labels-top    "M" "")
			   (if x-project-ticks "P" "")
			   (if x-major-ticks "T" "")
			   (if x-minor-ticks "S" "")
			   (if x-force-decimal "1" "")
			   (if x-force-exponential "2" "")))
	(yopt (concatenate 'string
			   (if y-axis "A" "")
			   (if y-left "B" "")
			   (if y-right "C" "")
			   (if y-grid "G" "")
			   (if y-invert-ticks "I" "")
			   (if y-log "L" "")
			   (if y-num-labels-left "N" "")
			   (if y-num-labels-right  "M" "")
			   (if y-project-ticks "P" "")
			   (if y-major-ticks "T" "")
			   (if y-minor-ticks "S" "")
			   (if y-vertical-num-labels "V" "")
			   (if y-force-decimal "1" "")
			   (if y-force-exponential "2" ""))))
    ;;
    (with-pgplot-settings (p :font font :character-height size :color color)
	   (pgbox xopt x-tick nxsub yopt y-tick nysub))
    (fix-pen-position p)
    t))


(defmethod plotlabel
    ((p pgplot) (xlabel string) (ylabel string) (toplabel string))
    "label pgplot device P with XLABEL, YLABEL, TOPLABEL.
Labels are not well placed, and (pgplot:xlabel p \"label\")
is preferred."
  (activate-device p)
  (pglab xlabel ylabel toplabel)
  (fix-pen-position p)
  t)

;; is m a 2x3 numerical matrix?  local to pgplot
(defun 2x3-numerical-matrix-p (m) 
  (and (arrayp m) (equalp '(2 3) (array-dimensions m)))
  (realp (aref m 0 0))  (realp (aref m 1 0))
  (realp (aref m 0 1))  (realp (aref m 1 1))
  (realp (aref m 0 2))  (realp (aref m 1 2)))

;; return min and max of 2d matrix a, ignoring values set to ignore-value
(defun 2d-array-min-max (a ignore-value)
  (let ((amin nil) (amax nil) (aa 0.0))
    (dotimes (i (array-dimension a 0))
      (dotimes (j (array-dimension a 1))
	(setf aa (aref a i j))
	(if (or (not ignore-value) (not (= aa ignore-value)))
	    (progn
	     (if (or (not amin) (< aa amin)) (setf amin aa)
	       (if (or (not amax) (> aa amax)) (setf amax aa)))))))
    (values amin amax)))
      
(declaim (inline to-single-float))
(defun to-single-float (x)
  (cond ((< x most-negative-single-float)
	 most-negative-single-float)
	((> x most-positive-single-float)
	 most-positive-single-float)
	;; we use  -normalized because plain least-postitive-float give
	;; underflows on some architectures (dangit..)
	((and (< x least-positive-normalized-single-float)
	      (> x least-negative-normalized-single-float))
	 0.0s0)
	(t
	 (coerce x 'single-float))))
(defun convert-array-to-single-float-array (a)
  (cond ((typep a '(simple-array single-float (* *)))
	 a)
	(t
	 (let* ((ni (array-dimension a 0))
		(nj (array-dimension a 1))
		(v  (make-array (list ni nj) :element-type 'single-float)))
	   (declare (type (unsigned-byte 28) ni nj))
	   (dotimes (i  ni)
	     (dotimes (j nj)
	       (declare (type (unsigned-byte 28) i j))
	       (setf (aref v i j) (to-single-float (aref a i j)))))
	   v))))

;; a=a(i)(j)
(defmethod contour ((p pgplot) (a array) &key
		    (contours nil)  ;; vector of contours
		    (n-contours 10) 
		    (min-contour nil) (max-contour nil)
		    (transformation-matrix nil)
		    (i-min-index nil) (i-max-index nil)
		    (j-min-index nil) (j-max-index nil)
		    (blank-value nil) ;; value of A at which contour undefined
		    (line-width nil) (color nil) (line-style nil)
		    (contour-labels nil);;vector of contour-labels, NIL ignored
		    (intval 20);;spacing of contour labels, in cells
		    (minint 10));;contours that cross <minint cells not labeled
  "create a contour plot of array A[i,j].  Contours are made at locations in
vector CONTOURS, but if this is NIL then N-CONTOURS evenly spaced contours
between optional MIN-CONTOUR and MAX-CONTOUR are made instead.
TRANSFORMATION-MATRIX (as described in pgplot docs) transforms array i,j to
user coordinates.  I-MIN-INDEX...etc are the ranges within A to use;
by default the whole array is used.   CONTOUR-LABELS is an optional
vector of contour labels corresponding to CONTOURS containing labels,
and INTVAL and MININT specify when to label contours, as described
in pgplot manual.  BLANK-VALUE is an optional out-of-range value which
should not be used when finding MIN/MAX of A[i,j].

For the data array A[I,J], the default coordinate system has
X increasing with J, and Y increasing with I, rescaled to fit
exactly within the user coordinates.   If used, the
transformation matrix (TR) changes this to:
  X = TR[0,0] + TR[0,1]*J + TR[0,2]*I
  Y = TR[1,0] + TR[1,1]*J + TR[1,2]*I
"
  (assert (= (array-rank a) 2))
  ;; fix  i-min-index, etc
  (let ((nx (+ -1 (array-dimension a 0)))
	(ny (+ -1 (array-dimension a 1))))
    (declare (type (unsigned-byte 28) nx ny))
    (if (not i-min-index) (setf i-min-index 0))
    (if (not j-min-index) (setf j-min-index 0))
    (if (not i-max-index) (setf i-max-index nx)) 
    (if (not j-max-index) (setf j-max-index ny))
    (if (or (< i-min-index 0) (> i-min-index nx)
	    (< j-min-index 0) (> j-min-index ny))
	(error "contour: bad min/max indices")))

  (if (and transformation-matrix
	   (not (2x3-numerical-matrix-p a)))
      (error "contour: transformation matrix ~A not a 2x3 matrix" 
	     transformation-matrix))
  ;;
  (let ((tv (make-array 6 :element-type 'single-float))
	;; convert array NOW because we may need to convert it many times
	(af (convert-array-to-single-float-array a)) 
	(cvec nil)    ;; vector of contours
	(minc nil)    ;; min contour
	(maxc nil))   ;; max contour
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; build transformation vector 
    (if (not transformation-matrix)
	;; if the transformation matrix is undefined, make the
	;; array fit into the current coordinate box
	(let* ((xmin (first   (get-window p)))
	       (xmax (second  (get-window p)))
	       (ymin (third   (get-window p)))
	       (ymax (fourth  (get-window p)))
	       (nx (array-dimension a 1)) ;; note flip
	       (ny (array-dimension a 0)) ;;  of x,y
	       (dx (/  (- xmax xmin) (- nx 1.0)))
	       (dy (/  (- ymax ymin) (- ny 1.0))))
	  (setf (aref tv 0) (to-single-float (- xmin dx)))
	  (setf (aref tv 3) (to-single-float (- ymin dy)))
	  (setf (aref tv 1) (to-single-float dx))
	  (setf (aref tv 2) 0.0)
	  (setf (aref tv 4) 0.0)
	  (setf (aref tv 5) (to-single-float dy)))
        ;;
        ;; if tranformation matrix is defined, use it for tv
	(progn
	  ;; tv elements 0,3 need to be adjusted to take into acct fact that
	  ;; fortran arrays start at 1; eg offset should be -dx,-dy not 0,0
	  (setf (aref tv 0) (to-single-float   ;; x0-adj
			     (- (aref transformation-matrix 0 0)    ;; x0
				(aref transformation-matrix 0 1)))) ;; dx
	  (setf (aref tv 3) (to-single-float   ;; y0-adj
			     (- (aref transformation-matrix 1 0)    ;; y0
				(aref transformation-matrix 1 2)))) ;; dy
	  (setf (aref tv 1) (to-single-float (aref transformation-matrix 0 1)))
	  (setf (aref tv 2) (to-single-float (aref transformation-matrix 0 2)))
	  (setf (aref tv 4) (to-single-float (aref transformation-matrix 1 1)))
	  (setf (aref tv 5) (to-single-float (aref transformation-matrix 1 2)))
	  ))
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; now compute contour levels
    (if contours
	(progn
	  (setf cvec contours)  ;; contour vector given, so ignore all else
	  (setf n-contours (length contours)))
      (progn ;; if contours not given, compute them
	(setf minc min-contour)
	(setf maxc max-contour)
	;; if min-contour and max-contour were not defined, then use A
	(if (not (and minc maxc))
	    (multiple-value-bind (mina maxa) (2d-array-min-max a blank-value)
	      (if (not minc) (setf minc mina))
	      (if (not maxc) (setf maxc maxa))))
	;;
	(if (not (and minc maxc))
	    (error
	     (concatenate
	      'string
	      "contour: could not find min/max elements in array"
	      " maybe all the array elements are the blank-value?")))
	;;
	(setf minc (to-single-float minc))
	(setf maxc (to-single-float maxc))
	(setf cvec (make-array n-contours :element-type 'single-float))
	(dotimes (i n-contours)
	  (declare (type (unsigned-byte 28) i))
	  (setf (aref cvec i)
		(to-single-float
		 (+ minc (* 1.0s0 (/ (- maxc minc) (- n-contours 1.0s0))  i))
		 )))))
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; do the contour plot
    (activate-device p)
    (with-pgplot-settings (p :line-width line-width :color color
			     :line-style line-style)
      (pgcont af
	      i-min-index i-max-index
	      j-min-index j-max-index
	      cvec (- n-contours) ;; minus for solid lines
	      tv)
      ;;
      ;; now do contour-labels
      (when contour-labels
	(loop for contour across cvec
	      for contour-label across contour-labels
	      for contour-label-string = (if (stringp contour-label)
					     contour-label
					     (format nil "~A" contour-label))
	      do
	      (when contour-label ;; NIL labels are ignored
		(pgconl af
			i-min-index i-max-index
			j-min-index j-max-index
			contour
			tv 
			contour-label-string 
			intval minint)))))
    (fix-pen-position p)
    t)) 




(defmethod connect ((p pgplot) (xvec vector) (yvec vector)
		    &key (line-width nil) (color nil) (line-style nil))
  "For pgplot device P, draw lines connecting pairs in XVEC, YVEC
with LINE-WIDTH (integer), COLOR (integer or symbol), and LINE-STYLE
 (integer or symbol)"
  (activate-device p)
  (with-pgplot-settings (p :line-width line-width :line-style line-style
			   :color color)
	  (pgline (min (length xvec) (length yvec)) xvec yvec))
  (fix-pen-position p)
  t)

(defmethod errorbars ((p pgplot) (xvec vector) (yvec vector) (evec vector)
		      &key (direction :plus-y) (size 1.0) (line-width nil)
		      (color nil) (line-style nil))
  "For pgplot device P, draw errorbars at points represented by pairs
in XVEC, YVEC, of magnitude of points in vector EVEC.
DIRECTION can be :PLUS-X :PLUS-Y :MINUS-X :MINUS-Y.
LINE-WIDTH is an integer, COLOR is an integer or symobl, and
LINE-STYLE is an integer or symbol."
  (activate-device p)
  (let ((n (min (length xvec) (length yvec) (length evec)))
	(ndir (case direction
		(:plus-x 1) (:plus-y 2) (:minus-x 3) (:minus-y 4))))
    (if (not ndir)
	(error
            "Bad direction in errorbar: allowed directions are :plus-x :plus-y :minus-x :minus-y"))
    (with-pgplot-settings (p :line-width line-width :line-style line-style
			     :color color)
      (pgerrb ndir n xvec yvec evec size))
    (fix-pen-position p)
    t))


(defmethod erase ((p pgplot))
  "Erase a pgplot device P"
  (activate-device p)
  (pgeras-raw)
  (fix-pen-position p)
  t)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar *point-symbol-alist*
  '((:OPEN-SQUARE . 0) (:DOT . 1) (:PLUS . 2) (:ASTERISK . 3) (:TIMES . 5)
    (:OPEN-TRIANGLE . 7) (:CIRCLE-PLUS . 8) (:CIRCLE-DOT . 9)
    (:SQUARE-WITH-CORNER-DOTS . 10) (:DIAMOND . 11) (:OPEN-STAR . 12)
    (:OPEN-PLUS . 14) (:STAR-OF-DAVID . 15) (:FILLED-SQUARE . 16)
    (:SPOT . 17) (:FILLED-STAR . 18) (:LARGER-SQUARE . 19)
    (:CIRCLE-1 . 20) (:CIRCLE-2 . 21) (:CIRCLE-3 . 22) (:CIRCLE-4 . 23)
    (:CIRCLE-5 . 24) (:CIRCLE-6 . 25) (:CIRCLE-7 . 26) (:CIRCLE-8 . 27)
    (:LEFT-ARROW . 28) (:RIGHT-ARROW . 29) (:UP-ARROW . 30) (:DOWN-ARROW . 31)
    (:POINT . -1) (:FILLED-TRIANGLE . -2) (:FILLED-DIAMOND . -4)
    (:FILLED-PENTAGON . -5) (:FILLED-HEXAGON . -6) (:FILLED-SEPTAGON . -7)
    (:FILLED-OCTAGON . -8)))
;;;;;;;;;;;; x,y are either vectors or reals

(defgeneric points ((p pgplot) x y symbol &key size  color)
  (:documentation
   "In pgplot device P draw SYMBOLS at X,Y (vector or real).
SYMBOL may be a symbol in pgplot::*point-symbol-alist* or
an integer.  SIZE is a real, and COLOR is an integer or symbol"))

(defmethod points ((p pgplot) x y (symbol integer)
		   &key (size nil) (color nil))
  (activate-device p)
  (if (or (> symbol 2932)
	  (< symbol -8))
      (error "Bad point symbol ~A.  Points must be in range -8 to 2132"
	     symbol))
  (with-pgplot-settings (p :character-height size :color color)
    (cond ((and (vectorp x) (vectorp y))
	   (pgpt (min (length x) (length y)) x y symbol))
	  ((and (realp x) (realp y))
	   (pgpt 1 (vector x) (vector y) symbol))
	  (t
	   (error "Invalid or mismatched x,y types in points"))))
  (fix-pen-position p)
  t)

(defmethod points ((p pgplot) x y (symbol symbol)
		   &key (size nil) (color nil))  
  (let ((ns (cdr (assoc symbol *point-symbol-alist*))))
    (if (not ns)
	(error
	 "Invalid point symbol ~A -- Available symbols are ~%~A ~%~A"
	  symbol (map 'list #'car  *point-symbol-alist*)
	  "Symbols may also be numbers 32-127 (ASCII) or 128-2132 (Hershey)"))
    (points p x y ns :size size :color color)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; convert world coordinates to device coordinates	
(defgeneric world-to-device  ((p pgplot) x y)
    (:documentation
     "for pgplot device P convert world coordinates X,Y (either
reals or vectors) to device coordinates in [0,1]"))   
	   
(defmethod world-to-device ((p pgplot) (x real) (y real))
  (let* ((vp (get-viewport p))
	 (w  (get-window   p))
	 (vx1 (first vp)) (vx2 (second vp)) (vy1 (third vp)) (vy2 (fourth vp))
	 (wx1 (first w))  (wx2 (second w))  (wy1 (third w))  (wy2 (fourth w)))
    (values
     (+ vx1 (* (/ (- x wx1) (- wx2 wx1)) (- vx2 vx1)))
     (+ vy1 (* (/ (- y wy1) (- wy2 wy1)) (- vy2 vy1))))))

;; for vectors
(defmethod world-to-device ((p pgplot) (x vector) (y vector))
  (if (not (= (length x) (length y)))
      (error "Unequal vectors in  world-to-device"))
  (let ((xx 0.0) (yy 0.0)
	(xv (make-array (length x) :element-type 'single-float))
	(yv (make-array (length x) :element-type 'single-float)))
    (dotimes (i (length x))
      (multiple-value-setq (xx yy) (world-to-device p (aref x i) (aref y i)))
      (setf (aref xv i) (to-single-float xx))
      (setf (aref yv i) (to-single-float yy)))
    (values xv yv)))

;; convert device coordinates to world coordinates
(defgeneric device-to-world  ((p pgplot) x y)
    (:documentation
     "for pgplot device P convert device coordinates X,Y 
reals or vectors) to user coordinates"))   
	   
(defmethod device-to-world ((p pgplot) (x real) (y real))
  (let* ((vp (get-viewport p))
	 (w  (get-window   p))
	 (vx1 (first vp)) (vx2 (second vp)) (vy1 (third vp)) (vy2 (fourth vp))
	 (wx1 (first w))  (wx2 (second w))  (wy1 (third w))  (wy2 (fourth w)))
    (values
     (+ wx1 (* (/ (- x vx1) (- vx2 vx1)) (- wx2 wx1)))
     (+ wy1 (* (/ (- y vy1) (- vy2 vy1)) (- wy2 wy1))))))

;; this time for vectors
(defmethod device-to-world ((p pgplot) (x vector) (y vector))
  (if (not (= (length x) (length y)))
      (error "Unequal vectors in  device-to-world"))
  (let ((xx 0.0) (yy 0.0)
	(xv (make-array (length x) :element-type 'single-float))
	(yv (make-array (length x) :element-type 'single-float)))
    (dotimes (i (length x))
      (multiple-value-setq (xx yy) (device-to-world p (aref x i) (aref y i)))
      (setf (aref xv i) xx)
      (setf (aref yv i) yy))
    (values xv yv)))
    

(defmethod write-text ((p pgplot) (text string) (x real) (y real) &key
		       (fjust 0.0) (angle 0.0) (device-coords nil)
		       (color nil) (character-height nil) (font nil))
  "For pgplot device P, write text string STRING at location x,y,
FJUST is justification (0=left justification at x,y; 1=right justifid).
ANGLE is angle counterclockwise from horizontal.
COLOR is a symbol or integer;  CHARACTER-HEIGHT is a real;
FONT is a symbol or integer"
  ;; if device coords non-nil, assume x,y are really in device coordinates
  ;; and transform to world coords
  (if device-coords (multiple-value-setq (x y) (device-to-world p x y)))
  (activate-device p)
  (with-pgplot-settings (p :color color :character-height character-height
			   :font font)
    (pgptxt x y angle fjust text))
  (fix-pen-position p)
  t)
  
;; return width and height of string in units where plot has width 1.
;(defmethod string-size ((p pgplot) (text string))
;  (multiple-value-bind (xlen ylen) (pglen 2 text)
;    (values (/ xlen (get-width p))
;	    (/ (* (get-aspect-ratio p) ylen (get-width p))))))

;; return two vectors representing bounding box of text
(defmethod text-bounding-box  ((p pgplot) (text string) 
			       &key (x0 0.0) (y0 0.0) (angle 0.0) (fjust 0.0)
			       (device-coords nil)
			       (character-height nil) (font nil))
  "Like WRITE-TEXT but returns two vectors representing the bounding
box the text would occupy if printed"
  (if (or (eq device-coords t) (eq device-coords :in))
      (multiple-value-setq (x0 y0) (device-to-world p x0 y0)))
  (activate-device p)
  (let ((xvec (make-array 4 :element-type 'single-float))
	(yvec (make-array 4 :element-type 'single-float)))
    (with-pgplot-settings (p :character-height character-height :font font)
      (pgqtxt x0 y0 angle fjust text xvec yvec))
    (if (or (eq device-coords t) (eq device-coords :out))
	(world-to-device p xvec yvec) ;; back to device coords
	(values xvec yvec))))
      

(defmethod xlabel  ((p pgplot) (text string) &key
		    (x-offset 0.0) (y-offset 0.0)
		    (color nil) (character-height nil) (font nil))
  "For pgplot device P, draw x axis label TEXT, nudging it
by X-OFFSET, Y-OFFSET in device coordinates.
If used, COLOR is an integer or symbol, CHARACTER-HEIGHT is a real,
and FONT is an integer or symbol"
  (activate-device p)
  (let ((theight 0.0)  ;; text width
	(twidth  0.0)
	(ybot    (third  (get-viewport p)))
	(xleft   (first  (get-viewport p)))
	(xright  (second (get-viewport p))))
    (multiple-value-bind (xbox ybox)
	(text-bounding-box p text :device-coords t
			   :character-height character-height :font font)
      (setf twidth  (abs (- (aref xbox 2) (aref xbox 0))))
      (setf theight (abs (- (aref ybox 2) (aref ybox 0)))))
    ;;
    (write-text p text
		(+ x-offset xleft (* 0.5 (- xright xleft twidth)))
		;; the 0.05 takes into account a guess at the axis labels
		(max (+ y-offset  (- ybot 0.05 (* 1.0 theight)))
		     0)
		:device-coords t
		:color color :character-height character-height :font font)
    t))

(defmethod ylabel  ((p pgplot) (text string) &key (x-offset 0.0) (y-offset 0.0)
		    (color nil) (character-height nil) (font nil))
    "For pgplot device P, draw y axis label TEXT, nudging it
by X-OFFSET, Y-OFFSET in device coordinates.
If used, COLOR is an integer or symbol, CHARACTER-HEIGHT is a real,
and FONT is an integer or symbol"
  (activate-device p)
  (let ((theight 0.0)  ;; text width
	(twidth  0.0)
	(ybot    (third  (get-viewport p)))
	(ytop    (fourth (get-viewport p)))
	(xleft   (first  (get-viewport p))))
    (multiple-value-bind (xbox ybox)
	(text-bounding-box p text :device-coords t :angle 90.0
			   :character-height character-height :font font)  
      (setf twidth  (abs (- (aref xbox 2) (aref xbox 0))))
      (setf theight (abs (- (aref ybox 2) (aref ybox 0)))))
      ;;
      (write-text p text
		  (max (+ x-offset (- xleft 0.05 (* 1.0 twidth)))
		       0)
		  (+ y-offset ybot (* 0.5 (- ytop ybot theight)))
		  :angle 90.0 :device-coords t
		  :color color :character-height character-height :font font)
      t))

		        
(defmethod toplabel ((p pgplot) (text string) &key (x-offset 0.0)
		     (y-offset 0.0) (color nil) (character-height nil)
		     (font nil))
    "For pgplot device P, draw top label TEXT, nudging it
by X-OFFSET, Y-OFFSET in device coordinates.
If used, COLOR is an integer or symbol, CHARACTER-HEIGHT is a real,
and FONT is an integer or symbol"
  (activate-device p)
  (let ((theight 0.0)  ;; text width
	(twidth  0.0)
	(ytop    (fourth (get-viewport p)))
	(xleft   (first  (get-viewport p)))
	(xright  (second (get-viewport p))))
    (multiple-value-bind (xbox ybox)
	(text-bounding-box p text :device-coords t
			   :character-height character-height :font font)
      (setf twidth  (abs (- (aref xbox 2) (aref xbox 0))))
      (setf theight (abs (- (aref ybox 2) (aref ybox 0)))))
    ;;
    (write-text p text
		(+ x-offset xleft (* 0.5 (- xright xleft twidth)))
		(+ y-offset  (+ ytop 0.03 ))
		:device-coords t
		:color color :character-height character-height :font font)
    t))
    
  
(defmethod get-cursor-position ((p pgplot) &key (x nil) (y nil)
				(use-last-pos t))
  "For pgplot device P, get the cursor position, returning
 (values x y char-number), initially placing cursor at keyword arguments
x,y or using the last cursor position if use-last-pos is T, the default" 
  (when use-last-pos
    (setf x (or x (get-last-curs-x p)))
    (setf y (or y (get-last-curs-y p))))
  (when (not x) (setf x (first (get-pen-position p))))
  (when (not y) (setf y (second (get-pen-position p))))
  (activate-device p)
  (multiple-value-bind 
   (xx yy char) (pgcurs x y)
   (setf (get-last-curs-x p) xx)
   (setf (get-last-curs-y p) yy)
   (values xx yy char)))


(defmethod draw-arrow ((p pgplot)
		       (x1 real) (y1 real) (x2 real) (y2 real)
		       &key (size nil) (color nil))
  "For pgplot device P, draw an arrow from X1,Y1 to X2,Y2, using
current arrow qualities.  SIZE is a real."
  (activate-device p)
  (with-pgplot-settings (p :character-height size :color color)
    (pgarro x1 y1 x2 y2))
  (fix-pen-position p)
  t)


(defmethod circle ((p pgplot) (x0 real) (y0 real) (radius real)
		   &key (line-width nil) (color nil) (line-style nil)
		   (fill-area-style nil))
  "For pgplot device P, draw a circle of radius RADIUS at X0,Y0
LINE-WIDTH is an integer, COLOR an integer or symbol, LINE-STYLE an
integer or symbol, and FILL-AREA-STYLE an integer or symbol."
    (activate-device p)
    (with-pgplot-settings (p :line-width line-width
			     :color color
			     :line-style line-style
			     :fill-area-style fill-area-style)
      (pgcirc x0 y0 radius))
    (fix-pen-position p)
    t)


;; theta-deg is counterclockwise angle of major axis to positive x axis
(defmethod ellipse ((p pgplot) (x-center real) (y-center real)
		    (major-axis real) (minor-axis real)
		    (theta-deg real)
		    &key (line-width nil) (color nil) (line-style nil)
		    (fill-area-style nil) (n-points 360))
    "For pgplot device P, draw an ellipse of semi-major-axes
MAJOR-AXIS and MINOR-AXIS, with the major axis at an angle
THETA-DEG counterclockwise from +x axis.
LINE-WIDTH is an integer, COLOR an integer or symbol, LINE-STYLE an
integer or symbol, and FILL-AREA-STYLE an integer or symbol.
N-POINTS is the number of segments to use."
  (when (> (abs theta-deg) 65536)
    (error "Out of range theta-deg - should have |theta-deg|<65536"))
  (let* ((theta-deg (the (float -65536.0 65536.0)
		      (coerce theta-deg 'single-float)))
	 (xvec (make-array n-points :element-type 'single-float))
	 (yvec (make-array n-points :element-type 'single-float))
	 (cos-rot (cos  (* 0.017453292 theta-deg)))
	 (sin-rot (sin  (* 0.017453292 theta-deg)))
	 (xx 0.0)
	 (yy 0.0)
	 (x  0.0)
	 (y  0.0)
	 (phi 0.0))
    (declare (type (single-float -10.0 10.0) phi)) ;; make trig functions happy
    (dotimes (i n-points)
      (setf phi (* (/ (float i) (float n-points))  6.28318530717s0))
      (setf xx (* (cos phi) major-axis))
      (setf yy (* (sin phi) minor-axis))
      (setf x (+ (* xx cos-rot) (* -1.0 yy sin-rot)))
      (setf y (+ (* xx sin-rot) (* yy cos-rot)))
      (setf (aref xvec i) (to-single-float (+ x x-center) ))
      (setf (aref yvec i) (to-single-float (+ y y-center) )))
    (polygon p xvec yvec
	     :line-width line-width :color color :line-style line-style
	     :fill-area-style fill-area-style)))



(defmethod rectangle ((p pgplot) (x1 real) (x2 real) (y1 real)
		      (y2 real)
		      &key (line-width nil) (color nil) (line-style nil)
		      (fill-area-style nil))
      "For pgplot device P, draw an rectangle with corners X1,Y1 and X2,Y2
LINE-WIDTH is an integer, COLOR an integer or symbol, LINE-STYLE an
integer or symbol, and FILL-AREA-STYLE an integer or symbol."
    (activate-device p)
    (with-pgplot-settings (p :line-width line-width
			     :color color
			     :line-style line-style
			     :fill-area-style fill-area-style)
      (pgrect x1 x2 y1 y2))
    (fix-pen-position p)
    t)

(defmethod polygon ((p pgplot) (xvec vector) (yvec vector) 
		   &key (line-width nil) (color nil) (line-style nil)
		   (fill-area-style nil))
        "For pgplot device P, draw a polygon with vertices
given by pairs in vectors XVEC YVEC.
LINE-WIDTH is an integer, COLOR an integer or symbol, LINE-STYLE an
integer or symbol, and FILL-AREA-STYLE an integer or symbol."
    (activate-device p)
    (with-pgplot-settings (p :line-width line-width
			     :color color
			     :line-style line-style
			     :fill-area-style fill-area-style)
      (pgpoly (min (length xvec) (length yvec)) xvec yvec))
    (fix-pen-position p)
    t)


;; sets background color, destructively modifying colormap
;; COLORMAP IS NO LONGER VALID (should we insert new name too?)
(defmethod set-colormap-color ((p pgplot) (ic integer) 
			       (r real) (g real) (b real)
			       &key (name nil))
  "For pgplot device P, replace the current colormap's color
denoted by integer IC with one given by the reals R,G,B
NAME is the optional new name for the color"
  (ensure-type 'single-float r g b)
  (if (or (< r 0) (> r 1) (< g 0) (> g 1) (< b 0) (> b 1))
      (error "Invalid colors in pgplot:set-background-color - must be in [0,1]"))
  (let* ((cm (get-colormap p))
	 (alist-new (remove-if #'(lambda (pair) (= (cdr pair) ic))
			       (colormap-color-alist cm))))
    (if (or (>= ic (colormap-npoints cm)) (< ic 0))
	(error "Invalid color index in set-colormap-color"))
    (if (member name alist-new)
	(print "WARNING: two colors will have same name in colormap"))
    ;; set up new color alist
    (if name 
	(setf (colormap-color-alist cm) (cons (cons name ic) alist-new))
      (setf (colormap-color-alist cm)  alist-new))
    ;;
    (setf (aref (colormap-red-vec   cm) ic) r)
    (setf (aref (colormap-red-vec   cm) ic) g)
    (setf (aref (colormap-red-vec   cm) ic) b))
  (pgscr ic r g b))
  


;; set current viewport to be a pane of larger plot
;; ix=1..nx is x number, iy=1...ny is y number
;; scales are in normalized device coords 0...1
(defmethod set-current-pane ((p pgplot)
			     (ix integer) (nx integer)
			     (iy integer) (ny integer)
			     &key (x-separation 0.05)
			     (y-separation 0.05)
			     (top-pad 0.02)    ;; extra blank space on top
			     (bottom-pad 0.10) ;; on bottom
			     (right-pad 0.03)  ;; on right
			     (left-pad 0.15))  ;; on left
  "For pgplot device P, divide the entire plottng area so that the 
current viewport correspond to the IX'th of NX and IY'th of NY
sub-panes.  X-SEPARATION, Y-SEPARATION are the distances to
assume between panes; TOP-PAD, BOTTOM-PAD, LEFT-PAD, RIGHT-PAD
are blank spaces to leave at sides of entire plot, to allow
for labels"
  (let* ((x-avail (- 1.0 left-pad right-pad))
	 (y-avail (- 1.0 top-pad bottom-pad))
	 (xsize (/ (- (float x-avail) (* (- nx 1.0) x-separation))
		  (float nx)))
	 (ysize (/ (- (float y-avail) (* (- ny 1.0) y-separation))
		  (float ny)))
	 (x0 (+ left-pad (* (- ix 1) xsize) (* (- ix 1) x-separation)))
	 (y0 (+ bottom-pad (* (- iy 1) ysize) (* (- iy 1) y-separation)))
	 (x1 (+ x0 xsize))
	 (y1 (+ y0 ysize)))
    (set-viewport p x0 x1 y0 y1)))

	
			     

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; histogram routines
;; to do: options for input binvec, equal-number-per-bin, hatching
;; arguments
;;   p       - the pgplot object
;;   xpts    - a vector of points to histogram
;;   line-width,color,line-style - what line to use
;;   binvec  - vector of bin edges - overrides nbins, and incompatible with
;;             non-nil xleft,xright
;;   nbins   - number of bins
;;   xleft,xright - x limits for histogram (if none supplied, use xpts)
;;   log      - logarithmic (base 10) bins
;;   normalize - normalize histogram so that part that falls within binvec
;;               has unit integral; done before taking log, if :log set.
;;               Points falling outside binvec are ignored
;;   flip-x   - flip x axis so that positive x bins are on left; not
;;              compatible with non-nil binvec
;;   y-frac   - fraction of y axis filled by histogram
;;   y-scale  - number by which to multiply y
;;   y-max    - maximum y -- overrides y-frac
;;   box      - draw a box
;;   set-window - set window to xleft,xright (which may be taken from binvec)
;;                     if nil, histogram is drawn using current window
;;                     if t, histogram is drawn using new window
;;                     if :restore histogram is drawn using new window, but old is restored
;;   fill, fill-color, fill-line-width, fill-line-style -- fill histogram? fill is
;;                    nil (no fill, default), or a fill style like :hatched
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmethod hist ((p pgplot) (xpts vector)  
		 &key
		 (binvec nil)
		 (nbins 20) (xleft nil) (xright nil) (flip-x nil) (log nil)
		 (normalize nil)
		 (y-frac 0.8) (y-scale 1.0) (y-max nil)
		 (set-window t)
		 (line-width nil) (color nil) (line-style nil)
		 (fill nil) (fill-color nil) (fill-line-width nil) 
		 (fill-line-style nil))
  "For pgplot device P, histogram a vector of points XPTS.
If BINVEC is specified, use it as the bin boundaries, else make
 NBINS evenly spaced bins.
XLEFT, XRIGHT are the optional X axis limits for the plot, set by
 default to be the limits of XPTS
FLIP-X is a flag about whether to mirror-reflect the x axis.
LOG is a flag whether to plot the log10 of the number of points in
 each bin.
NORMALIZE is a flag denoting whether the bin counts should be
 normalized so the integral of the distribution between XLEFT and XRIGHT
 is 1.
Y-FRAC is the fraction of the total height the highest bin should reach.
Y-SCALE is a rescaling for the bin contents.
SET-WINDOW is a flag (T by default) that controls whether the window
 is reset before plotting. SET-WINDOW=NIL is useful for multiple plots
LINE-WIDTH is an optional integer, COLOR an integer or symbol,
  LINE-STYLE an integer or symbol
FILL is a boolean flag whether to fill the histogram, and
  FILL-COLOR, FILL-LINE-WIDTH FILL-LINE-STYLE are like
  COLOR, LINE-WIDTH, LINE-STYLE"
  (if (and (or xleft xright) (not (and xleft xright)))
      (error "hist: you must specify both xleft and xright, or neither"))
  (if (and flip-x xleft xright (< xleft xright))
      (error "hist: your have flip-x turned on, but this is contradicted by xleft < xright"))
  (if (and binvec (or xleft xright))
      (error "hist: you have *both* binvec and xleft or xright specified"))
  (if (and binvec flip-x)
      (error "hist: you have *both* binvec and flip-x specified"))
  ;;
  ;; if a binvector was not supplied, then make it
  ;; else check the one supplied for monotonic increase
  (if (not binvec)
      (progn
	(if (not xleft) ;; set xleft and xright if not given
	    (progn (setf xleft (aref xpts 0))
		   (setf xright (aref xpts 0))
		   (dotimes (i (length xpts))
		     (setf xleft (min xleft (aref xpts i)))
		     (setf xright (max xright (aref xpts i))))))
	;; flip xleft,xright if asked
	(if flip-x (let ((tmp xleft)) (setf xleft xright) (setf xright tmp)))
	(setf binvec (hist-make-binvec nbins xleft xright))) ;; and make the binvec
    (progn
      (setf nbins (+ -1 (length binvec))) ;; ignore any supplied nbins
      (multiple-value-setq (xleft xright binvec)
	(hist-check-binvec binvec))))


  ;;
  (let ((nbinvec  (make-array nbins :element-type 'single-float
			      :initial-element 0.0s0))
	(x-vec (make-array (+ 2 (* 2 nbins))
			   :element-type 'single-float))
	(y-vec (make-array (+ 2 (* 2 nbins))
			   :element-type 'single-float))
	(y-min 0.0s0))
    ;; fill nbinvec
    (hist-fill-nbinvec xpts binvec nbinvec
		       (coerce xleft 'double-float) (coerce xright 'double-float))
    ;; if normalization sought, normalize histogram so that integral between
    ;; xleft and xright is 1.0
    (if normalize (hist-normalize-histogram binvec nbinvec))
    ;; fill the xy vecs
    (hist-make-xy binvec nbinvec x-vec y-vec y-scale)



    (if log ;; is histogram to be logarithmic?
	(progn
	  (dotimes (i (length y-vec))
		   (setf (aref y-vec i)
			 (let ((y (aref y-vec i)))
			   (if (= y 0.0s0)
			       -10000.0
			     (let ((tmp (log y 10.s0)))
			       (setf y-min (min y-min (- tmp 1.0s0)))
			       tmp)))))))
    
    (if set-window
	(progn
	  (if (not y-max)
	      (progn
		(setf y-max (aref y-vec 0))
		(dotimes (i (length y-vec)) (setf y-max (max y-max (aref y-vec i))))
		(setf y-max (/ y-max y-frac))))
	  (set-window p xleft xright y-min y-max)))
      
    ;; and connect the points	
    (connect p x-vec y-vec :color color :line-width line-width
	     :line-style line-style)
    (if fill
	(polygon p x-vec y-vec :line-width fill-line-width :color fill-color
		 :fill-area-style fill :line-style fill-line-style)))
  
  (fix-pen-position p)
  t)

;; make this double-float so that all comparisons work
(defun hist-make-binvec (nbins xleft xright)
  (let ((binvec (make-array (+ nbins 1) :element-type 'double-float)))
    (dotimes (i  (+ 1 nbins))
      (setf (aref binvec i)
	    (coerce (+ xleft (* i (/ (- xright xleft) nbins))) 'double-float)))
    ;; now we nudge the outside bins outward by a little so that xleft
    ;; and xright always fall inside the bin vector, even with
    ;; numerical roundoff
    (let ((nudge (* 1.0e-12 (- xright xleft))))
      (incf (aref binvec 0) (- nudge))
      (incf (aref binvec nbins) nudge))
    binvec))


;; check binvec for monotonic increase, and coerce it to double float
(defun hist-check-binvec (binvec)
  (if (not (vectorp binvec)) (error "binvec supplied to hist not a vector"))
  (if (> 1 (length binvec)) (error "binvec supplied to hist too short"))
  ;;
  ;; if binvec not double float, copy it to a double float vec
  (if (not (equal (type-of binvec) '(simple-array double-float)))
      (let ((bv (make-array (length binvec) :element-type 'double-float)))
	(dotimes (i (length binvec))
	  (setf (aref bv i) (coerce (aref binvec i) 'double-float)))
	(setf binvec bv)))
  ;;
  (let ((sgn (if (< (aref binvec 0) (aref binvec 1)) 1.0 -1.0)))
    (dotimes (i (+ -1 (length binvec)))
      (if (> (* sgn (aref binvec i)) (* sgn (aref binvec (+ i 1))))
	  (error
 "hist: binvec supplied to hist not monotonically increasing or decreasing")))
    (values (aref binvec 0) (aref binvec (+ -1 (length binvec))) binvec)))
  

;; normalize histogram so that parts within binvec integrate to 1
(defun hist-normalize-histogram (binvec nbinvec)
    (declare (type (simple-array double-float) binvec)
	     (type (simple-array single-float) nbinvec)
	     (optimize (speed 3) (safety 1)))
    (let ((integral (abs
		     (the double-float
		       (loop for i fixnum from 0 to (- (length nbinvec) 1)
			     with sum of-type double-float = 0.0d0
			     do
			     (incf sum
				   (* (aref nbinvec i)
				      (- (aref binvec (+ i 1))
					 (aref binvec i))))
			     finally (return sum))))))
      (if (= integral 0.0d0)
	  (error "hist: cannot normalize histogram with all empty bins"))
      (loop for i fixnum from 0 to (- (length nbinvec) 1)
	    do (setf (aref nbinvec i)
		     (coerce (/ (aref nbinvec i) integral) 'single-float)))))
	  
	  
    
    

(defun hist-hunt-bin (binvec x)
  (let ((ibot 0)
	(itop  (length binvec))
	(icent 0)
	(dir (> (aref binvec 1) (aref binvec 0))))
    (declare (fixnum ibot itop icent))
		       (loop
			(if (<= (- itop ibot) 1) (return ibot)) 
			(setf icent (floor (+ ibot itop) 2))
			(if (eq (> x (aref binvec icent))  dir)
			    (setf ibot icent)
			  (setf itop icent)))))
    


(defun to-dbl-float-vec (v)
  (declare (type vector v))
  (cond ((typep v '(simple-array double-float (*)))
	 v)
	(t
	 (loop with vv = (make-array (length v) :element-type 'double-float)
	       for i fixnum from 0 below (length v)
	       do (setf (aref vv i) (coerce (aref v i) 'double-float))
	       finally (return vv)))))

;; use bisection to fill nbinvec -- types are assured by calling function
(defun hist-fill-nbinvec (xpts binvec nbinvec xleft xright)
  (declare (type (simple-array double-float) binvec)
	   (type (simple-array single-float) nbinvec)
	   (type simple-array xpts)
	   (type double-float xleft xright))
  (let ((dir (> (aref binvec 1) (aref binvec 0)))
	(xpts (to-dbl-float-vec xpts)))
    (declare (type (simple-array double-float (*)) xpts)
	     (optimize (speed 3) (safety 1)))
    (flet ((find-bin (x)
		     (declare (type double-float x)
			      (values fixnum))
		     (let ((ibot 0)
			   (itop   (+ -1 (length binvec)))
			   (icent 0))
		       (declare (type (unsigned-byte 28) ibot itop icent))
		       (loop
			(if (<= (- itop ibot) 1) (return ibot)) 
			(setf icent (floor (+ ibot itop) 2))
			(if (eq (> x (aref binvec icent))  dir)
			    (setf ibot icent)
			  (setf itop icent))))))
      ;;
      (declare (inline find-bin))
      ;;
      (dotimes (i (length xpts))
	(declare (type (unsigned-byte 28) i))
	(if (and (eq (>= (aref xpts i) xleft) dir)
		 (eq (<= (aref xpts i) xright) dir))
            (let ((j (find-bin (coerce (aref xpts i) 'double-float))))
	      (if (or (< j 0) (>= j (length nbinvec)))
		  (error "Invalid bin in hist-fill-nbinvec; should not happen"))
              (setf (aref nbinvec j) (+ 1.0 (aref nbinvec j)))))))))

	     
(defun hist-make-xy (binvec nbinvec x-vec y-vec y-scale)
  (setf (aref x-vec 0) (to-single-float (aref binvec 0)))
  (setf (aref y-vec 0) 0.0s0)
  ;;
  (dotimes (i (length nbinvec))
    (setf (aref x-vec  (+ (* i 2) 1))
	  (to-single-float (aref binvec  (+ i 0))))
    (setf (aref y-vec  (+ (* i 2) 1))
	  (to-single-float (* y-scale (aref nbinvec (+ i 0))) ))
    (setf (aref x-vec  (+ (* i 2) 2))
	  (to-single-float (aref binvec  (+ i 1)) ))
    (setf (aref y-vec  (+ (* i 2) 2))
	  (to-single-float (* y-scale (aref nbinvec (+ i 0))) )))
  ;;
  (setf (aref x-vec (+ -1 (length x-vec)))
	(to-single-float (aref binvec (+ -1 (length binvec)))))
  (setf (aref y-vec (+ -1 (length y-vec)))  0.0s0))

  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; image routines
(defmethod image ((p pgplot) (a array) &key
		  ;; type refers to name in (colormap-named-ranges colormap)
		  (type :default)  
		  (transformation-matrix nil)
		  (i-min-index nil) (i-max-index nil)
		  (j-min-index nil) (j-max-index nil)
		  ;; values in a that map to extremes of colors
		  (val1 nil) (val2 nil)
		  ;; max and min color indices if this is not the default
		  ;; colormap
		  (min-color-index nil) (max-color-index nil)
		  ;; how do values scale to colormap?
		  (transfer-function :linear))
  "For a pgplot P, plot 2d real array A as an image.
TYPE is the name of the colormap to use (eg :DEFAULT :GRAY :COLOR :HEAT)
TRANSFORMATION-MATRIX is a matrix to transform array coordinates to
  user coordinates.
I-MIN-INDEX, etc are the inclusive indices to use in array A[I,J]
VAL1 and VAL2 are optional values in A to the extremes of the color range.
MIN-COLOR-INDEX and MAX-COLOR-INDEX are the color index range to use,
 overriding TYPE.
TRANSFER-FUNCTION is the transfter function of the array to the color index,
 one of :LINEAR :LOG :SQRT

For the data array A[I,J], the default coordinate system has
X increasing with J, and Y increasing with I, rescaled to fit
exactly within the user coordinates.   If used, the
transformation matrix (TR) changes this to:
  X = TR[0,0] + TR[0,1]*J + TR[0,2]*I
  Y = TR[1,0] + TR[1,1]*J + TR[1,2]*I
"
  ;;
  (assert (= (array-rank a) 2))
  (assert (member transfer-function '(:linear :log :sqrt)))
  ;;
  ;; either we are using the default colormap, or we must specify
  ;; the index range explicitly
  (let ((cm (get-colormap p))
	(cmrange nil))
    (cond
     ;; indices specified
     (min-color-index
      (when (not (and min-color-index max-color-index
		      (< min-color-index max-color-index)
		      (> min-color-index 0)
		      (< max-color-index (colormap-npoints cm))))
	(error "image: Bad min/max color-indices ~a ~a" min-color-index max-color-index)))
     ;; else using a named range inside the colormap
     (t
      (setf cmrange (cdr (assoc type (colormap-named-ranges cm) :test 'eq)))
      (when (not cmrange)
	(error "image: Could not find named colormap range ~A - available are ~A"
	       type (colormap-named-ranges cm)))
      (setf min-color-index  (first cmrange)
	    max-color-index  (second cmrange)))))

  ;; fix  i-min-index, etc - these are inclusive lisp (0..n-1) indices
  (let ((nx (+ -1 (array-dimension a 0)))
	(ny (+ -1 (array-dimension a 1))))
    (if (not i-min-index) (setf i-min-index 0))
    (if (not j-min-index) (setf j-min-index 0))
    (if (not i-max-index) (setf i-max-index nx)) 
    (if (not j-max-index) (setf j-max-index ny))
    (if (or (< i-min-index 0) (> i-min-index nx)
	    (< j-min-index 0) (> j-min-index ny))
	(error "image: bad min/max indices")))
  ;;
  ;; fix the min/max values val1 and val2
  (when (or (not val1) (not val2))
    (loop for i fixnum from 1 below (array-total-size a)
	  for val = (row-major-aref a i)
	  for min-val = (row-major-aref a 0) then (min min-val val)
	  for max-val = (row-major-aref a 0) then (max max-val val)
	  finally
	  (when (not val1) (setf val1 min-val))
	  (when (not val2) (setf val2 max-val))))
  ;;
  (setf val1 (coerce val1 'single-float)
	val2 (coerce val2 'single-float))
  ;;
  (if (and transformation-matrix
	   (not (2x3-numerical-matrix-p a)))
      (error "image: transformation matrix ~A not a 2x3 matrix" 
	     transformation-matrix))
  (let ((tv (make-array 6 :element-type 'single-float)))
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; build transformation vector 
    (if (not transformation-matrix)
	;; if the transformation matrix is undefined, make the
	;; array fit into the current coordinate box
	(let* ((xmin (first   (get-window p)))
	       (xmax (second  (get-window p)))
	       (ymin (third   (get-window p)))
	       (ymax (fourth  (get-window p)))
	       (nx (array-dimension a 1)) ;; note flip
	       (ny (array-dimension a 0)) ;;  of x,y
	       (dx (/  (- xmax xmin) (- nx 1.0)))
	       (dy (/  (- ymax ymin) (- ny 1.0))))
	  (setf (aref tv 0) (to-single-float (- xmin dx)))
	  (setf (aref tv 3) (to-single-float (- ymin dy)))
	  (setf (aref tv 1) (to-single-float dx))
	  (setf (aref tv 2) 0.0)
	  (setf (aref tv 4) 0.0)
	  (setf (aref tv 5) (to-single-float dy)))
	;;
	;; if tranformation matrix is defined, use it for tv
	(progn
	  ;; tv elements 0,3 need to be adjusted to take into acct fact that
	  ;; fortran arrays start at 1; eg offset should be -dx,-dy not 0,0
	  (setf (aref tv 0) (to-single-float   ;; x0-adj
			     (- (aref transformation-matrix 0 0)    ;; x0
				(aref transformation-matrix 0 1)))) ;; dx
	  (setf (aref tv 3) (to-single-float   ;; y0-adj
			     (- (aref transformation-matrix 1 0)    ;; y0
				(aref transformation-matrix 1 2)))) ;; dy
	  (setf (aref tv 1) (to-single-float (aref transformation-matrix 0 1)))
	  (setf (aref tv 2) (to-single-float (aref transformation-matrix 0 2)))
	  (setf (aref tv 4) (to-single-float (aref transformation-matrix 1 1)))
	  (setf (aref tv 5) (to-single-float (aref transformation-matrix 1 2)))
	  ))
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; make the image plot
    (activate-device p)
    (pgscir min-color-index max-color-index) ;; set color index range (should be set already?)
    (pgsitf (cond ((eq transfer-function :linear) 0) ;; set transfer function
		  ((eq transfer-function :log) 1)
		  ((eq transfer-function :sqrt) 2)))
    (pgimag i-min-index i-max-index j-min-index j-max-index
	    a val1 val2 tv)
    ))

(provide 'pgplot)


 






